from ase import __version__ as ase_version
from ase.utils import convert_string_to_fd
from ase.utils import search_current_git_hash
from ase.visualize import view
import numpy as np
import os
import time
import sys
import xespresso
import ase

class XLogger():
    """Class for handling all text output."""
    def __init__(self):
        self.verbose = False
        self._fd = None
        self.oldfd = None

    @property
    def fd(self):
        return self._fd

    @fd.setter
    def fd(self, fd):
        """Set the stream for text output.
        """
        if fd == self.oldfd:
            return
        self.oldfd = fd
        self._fd = convert_string_to_fd(fd)
        self.logo()
        self.header()

    def __call__(self, *args, **kwargs):
        flush = kwargs.pop('flush', False)
        print(*args, file=self._fd, **kwargs)
        if flush:
            self._fd.flush()

    def flush(self):
        self._fd.flush()


    def logo(self):
        self(' xespresso  ')

    def header(self):
        self()
        nodename, machine = os.uname()[1::3]
        self('User:  ', os.getenv('USER', '???') + '@' + nodename)
        self('Date:  ', time.asctime())
        self('Arch:  ', machine)
        self('Pid:   ', os.getpid())
        self('Python: {0}.{1}.{2}'.format(*sys.version_info[:3]))
        # Espresso
        line = os.path.dirname(xespresso.__file__)
        githash = search_current_git_hash(xespresso)
        if githash is not None:
            line += ' ({:.10})'.format(githash)
        self('xespresso:  ', line)

        # ASE
        line = '%s (version %s' % (os.path.dirname(ase.__file__), ase_version)
        githash = search_current_git_hash(ase)
        if githash is not None:
            line += '-{:.10}'.format(githash)
        line += ')'
        self('ase:   ', line)
        self('\n'*2)

    def print_atoms(self, atoms):
        self()
        self('{0:15s}: {1}'.format('    Formula', atoms.get_chemical_formula()))
        self('{0:15s}: {1}'.format('    Cell', np.round(np.asarray(atoms.cell.cellpar()), 3)))
        self('{0:15s}: {1}'.format('    PBC', atoms.pbc))
        self('{0:15s}:'.format('    Info'))
        self.print_dict(atoms.info, sep = '        ')

    def print_calculator(self, calc):
        self()
        self('Calculator:')
        self('Package: ')
        self('  k-point: ', )
        self('  Pseudopotentials: \n')
        
    def print_dict(self, dct, sep='  '):
        options = np.get_printoptions()
        try:
            np.set_printoptions(threshold=4, linewidth=50)
            for key, value in sorted(dct.items()):
                if isinstance(value, dict):
                    sep = ',\n     ' + ' ' * len(key)
                    keys = sorted(value, key=lambda k: (str(type(k)), k))
                    s = sep.join('{0}: {1}'.format(k, value[k]) for k in keys)
                    self('  {0}: {{{1}}}'.format(key, s))
                elif hasattr(value, '__len__'):
                    value = np.asarray(value)
                    sep = ',\n    ' + ' ' * len(key)
                    s = sep.join(str(value).splitlines())
                    self('  {0}: {1}'.format(key, s))
                else:
                    self('  {0}: {1}'.format(key, value))
        finally:
            np.set_printoptions(**options)


from math import pi, sqrt
import numpy as np
from ase.dft.kpoints import get_monkhorst_pack_size_and_offset
from ase.parallel import world
from ase.utils.cext import cextension
from xespresso import Espresso
from ase.geometry import get_layers
from ase.visualize import view
import copy
import matplotlib.pyplot as plt
from ase.data.colors import jmol_colors
from ase.data import covalent_radii, atomic_numbers, chemical_symbols

lspins = ['up', 'dw']
orbitals = ['s', 'p', 'd', 'f']



class DOS:
    def __init__(self, calc = None, label = None, prefix = None,
                 debug = False, colors = None):
        """Electronic Density Of States object.

        calc: quantum espresso calculator object
        label: the directory and prefix

        dos_energies
        print(pdos_info) to see the detail
        pdos_kinds[species][channel][ncomponents]

        """
        if not calc and not label:
            raise ValueError('Please give one of them: calc or label')
        if label:
            calc = Espresso(label = label, prefix = prefix, debug = True)
        calc.read_results()
        self.debug = debug
        if self.debug:
            print(calc.results)
        self.directory = calc.directory
        self.label = calc.label
        self.prefix = calc.prefix
        self.efermi = calc.get_fermi_level()
        self.atoms = calc.results['atoms']
        self.nspins = calc.get_number_of_spins()
        if self.nspins == 0:  self.nspins = 1
    def read_dos(self):
        dos = np.loadtxt(self.directory + '/dos/%s.dos' % self.prefix)
        if self.nspins == 2:
            self.dos = [dos[:,1], dos[:, 2]]
        else:
            self.dos = [dos[:,1]]
        self.dos_energies = dos[:,0] - self.efermi
        # self.dos_energies = self.dos_energies.reshape((self.dos_energies.shape[0], 1))
    def read_pdos_info(self, ):
        '''
        read pdos information from output of projwfc
        state #   1: atom   1 (Al ), wfc  1 (l=0 m= 1)
        state #   2: atom   1 (Al ), wfc  2 (l=1 m= 1)
        state #   3: atom   1 (Al ), wfc  2 (l=1 m= 2)
        state #   4: atom   1 (Al ), wfc  2 (l=1 m= 3)
        state #   5: atom   2 (Al ), wfc  1 (l=0 m= 1)
        state #   6: atom   2 (Al ), wfc  2 (l=1 m= 1)
        state #   7: atom   2 (Al ), wfc  2 (l=1 m= 2)
        ...

        reg
        ['state', '1', 'atom', '1', 'Al', 'wfc', '1', 'l', '0', 'm', '1']
        Output: 
        post_info: dict,
            e.g. {'Al': {'iatom': [1, 2, 3, 4], 'istate': {1: 0, 2: 1}}}
        '''
        import re
        reg = re.compile('\w+')
        # 
        kinds = []
        pdos_info = {}
        with open(self.directory + '/projwfc/%s.projwfco' % self.prefix, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if 'state #' in line:
                    data = reg.findall(line)
                    iatom = int(data[3])
                    kind = data[4]
                    istate = int(data[6])
                    l = int(data[8])
                    if kind not in kinds:
                        kinds.append(kind)
                        pdos_info[kind] = {}
                        pdos_info[kind]['iatom'] = []
                        pdos_info[kind]['istate'] = {}
                    if iatom not in pdos_info[kind]['iatom']:
                        pdos_info[kind]['iatom'].append(iatom)
                    if istate not in pdos_info[kind]['istate']:
                        pdos_info[kind]['istate'][istate] = l
        self.kinds = kinds
        self.pdos_info = pdos_info
        if self.debug: print(self.pdos_info)
    def read_pdos(self):
        '''
        projwfc
        '''
        self.read_pdos_info()
        # read in total density of states
        dos = np.loadtxt(self.directory + '/projwfc/%s.pdos_tot' % self.prefix)
        if self.nspins == 2:
            self.pdos_tot = [dos[:,1], dos[:, 2]]
        else:
            self.pdos_tot = [dos[:,1]]
        self.pdos_energies = dos[:,0] - self.efermi
        npoints = len(self.pdos_energies)
        # read in projections onto atomic orbitals
        self.pdos_atoms = {}
        self.pdos_kinds = {}
        for kind, info in self.pdos_info.items():
            pdos_kind = {}
            for istate, l in info['istate'].items():
                ncomponents = (2*l+1) * self.nspins + 1
                channel = '{0}{1}'.format(istate, orbitals[l])
                pdos_kind[channel] = np.zeros((ncomponents, npoints), np.float)
            for iatom in info['iatom']:
                pdos_atom = {}
                for istate, l in info['istate'].items():
                    filename = self.directory + '/projwfc/{0}.pdos_atm#{1}({2})_wfc#{3}({4})'.format(self.prefix, iatom, kind, istate, orbitals[l])
                    channel = '{0}{1}'.format(istate, orbitals[l])
                    pdosinp = np.genfromtxt(filename)
                    ncomponents = (2*l+1) * self.nspins + 1
                    pdos_atom[channel] = np.zeros((ncomponents, npoints), np.float)
                    for j in range(ncomponents):
                        pdos_atom[channel][j] += pdosinp[:, j + 1]
                         # sum over kind
                self.pdos_atoms[iatom] = pdos_atom
            self.pdos_kinds[kind] = pdos_kind
        self.pdos_kinds = self.merge_kinds(index = range(len(self.atoms)))
        # sum over orbital
        return self.pdos_energies, self.pdos_tot, self.pdos_atoms, self.pdos_kinds
    def merge_kinds(self, index = []):
        '''
        merge atoms in the index
        '''
        pdos_kinds = {}
        npoints = len(self.pdos_energies)
        for kind, info in self.pdos_info.items():
            # print(kind, info)
            pdos_kind = self.pdos_kinds[kind]
            for channel, pdos in pdos_kind.items():
                # print(channel)
                pdos[:, :] = 0
            for iatom in info['iatom']:
                if iatom - 1 not in index: continue
                # print(iatom)
                for channel, pdos_atom in self.pdos_atoms[iatom].items():
                    pdos_kind[channel] += pdos_atom
            pdos_kinds[kind] = pdos_kind
        return pdos_kinds
    def read_proj(self, ):
        '''
        read projwfc output file *.up and *.dw
        '''
        #
        self.read_pdos_info()
        proj_files = [self.directory + '/projwfc/%s.projwfc_up'%self.prefix, self.directory + '/projwfc/%s.projwfc_down'%self.prefix]
        projs = []
        projs_kinds = []
        channels = []
        for i in range(self.nspins):
            with open(proj_files[i]) as f:
                lines = f.readlines()
                for j in range(len(lines)):
                    if '       1       1' in lines[j]: break
                nstates = int(lines[j - 3].split()[0])
                nkpts = int(lines[j - 3].split()[1])
                nbands = int(lines[j - 3].split()[2])
                print(nstates, nkpts, nbands)
                proj = np.zeros((nstates, nkpts, nbands), np.float)
                proj_kinds = {}
                for istate in range(nstates):
                    iline = j + istate*(nkpts*nbands + 1) - 1
                    kinds = lines[iline].split()[2]
                    orbital = lines[iline].split()[3]
                    channel = kinds + orbital
                    if channel not in channels: 
                        channels.append(channel)
                        proj_kinds[channel] = np.zeros((nkpts, nbands), np.float)
                    # l = lines[iline].split()[2]
                    # print(kinds, istate)
                    for ikpt in range(nkpts):
                        for iband in range(nbands):
                            iline = j + istate*(nkpts*nbands + 1) + ikpt*nbands + iband
                            # print('line: ', iline)
                            proj[istate, ikpt, iband] = float(lines[iline].split()[2])
                            proj_kinds[channel][ikpt, iband] += float(lines[iline].split()[2])
                            # proj_atoms[channel, ikpt, iband] += 
            projs.append(proj)
            projs_kinds.append(proj_kinds)
        self.projs = projs
        self.projs_kinds = projs_kinds
    def plot_data(self, energies, dos, label, xindex, ax = None, fill = True, color = None, smearing = None):
        if ax is None:
            ax = plt.gca()
        for i in range(self.nspins):
            if not smearing:
                xmesh = energies[xindex]
                ymesh = dos[i][xindex]
            else:
                xmesh, ymesh = self.smearing(energies[xindex], dos[i][xindex], sigma=smearing[0], de = smearing[0])
            ymesh = ymesh.reshape(-1, )
            if self.nspins == 2:
                newlabel = '%s-%s' % (label, lspins[i])
            else:
                newlabel = label
            if color is not None:
                ax.plot(xmesh, (-1)**i*ymesh, linewidth=0.7, label = newlabel, color = color)
            else:
                ax.plot(xmesh, (-1)**i*ymesh, linewidth=0.7, label = newlabel)
            if fill:
                if color is not None:
                    ax.fill_between(xmesh, (-1)**i*ymesh, 0, alpha = 0.2, color = color)
                else:
                    ax.fill_between(xmesh, (-1)**i*ymesh, 0, alpha = 0.2,)
        return ax
    def plot_dos(self, energies = None, dos = None, 
        Emin = -5, Emax = 5, color = None,
        ax = None, fill = True, output = None, smearing = None):
        '''
        '''
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots(figsize = (6, 3))
            # ax = plt.gca()
        if dos is None: dos = self.dos
        if energies is None: energies = self.dos_energies
        xindex = (energies>Emin) & (energies<Emax)
        self.plot_data(energies, dos, label = 'dos', ax = ax, xindex = xindex, fill = fill, color = color, smearing = smearing)
        # ax.legend()
        # plt.grid(linestyle = '--')
        # ax.set_xlabel('Energy (eV)')
        # ax.set_ylabel('DOS (a.u.)')
        # ax.set_title('%s' % self.prefix)
        if output is not None:
            plt.savefig('%s'%output)
        return ax
    def plot_pdos_tot(self, energies = None, dos = None, 
        Emin = -5, Emax = 5,
        ax = None, fill = True, output = None, smearing = None):
        # print(self.pdos_tot)
        if dos is None: dos = self.pdos_tot
        if energies is None: energies = self.pdos_energies
        xindex = (energies>Emin) & (energies<Emax)
        fig, ax = plt.subplots(figsize = (6, 3))
        self.plot_data(energies, dos, label = 'pdos', ax = ax, xindex = xindex, fill = fill, smearing = smearing)
        ax.legend()
        ax.set_xlabel('Energy (eV)')
        ax.set_ylabel('PDOS (a.u.)')
        ax.set_title('%s' % self.prefix)
        if output is not None:
            plt.savefig('%s'%output)
        return ax
    def plot_pdos(self, energies = None, pdos_kinds = None, 
                  Emin = -5, Emax = 5,
                  ax = None, total = False, select = None, fill = True, 
                  output = None, legend = False, xylabel = True, smearing = None):
        '''
        '''
        if energies is None: energies = self.pdos_energies
        if pdos_kinds is None: pdos_kinds = self.pdos_kinds
        xindex = (energies>Emin) & (energies<Emax)
        # print(xindex)
        if ax is None:
            fig, ax = plt.subplots(figsize = (6, 3))
        # if total:
            # self.plot_data(self.pdos_energies, self.pdos_tot, label = 'pdos', ax = ax, fill = fill)
        for kind, channels in pdos_kinds.items():
            if select and kind not in select: continue
            number = chemical_symbols.index(kind)
            color = jmol_colors[number]
            for channel, pdos in channels.items():
                if select and channel[-1] not in select[kind]: continue
                label = '{0}-{1}'.format(kind, channel)
                self.plot_data(energies, pdos, label = label, ax = ax, xindex = xindex, fill = fill, color = color, smearing = smearing)
        if legend: ax.legend()
        if xylabel:
            ax.set_xlabel('Energy (eV)')
            ax.set_ylabel('PDOS (a.u.)')
        if output is not None:
            plt.savefig('%s'%output)
        return ax
    def plot_pdos_layer(self, atoms = None, 
                        layers = None,
                        miller = (0, 0, 1), 
                        dz = 0.3, 
                        Emin = -5, 
                        Emax = 3, 
                        total = False, 
                        select = None, 
                        fill = True, 
                        smearing = None,
                        output = None):
        '''
        
        '''
        atoms = self.atoms
        if not layers:
            layers = get_layers(atoms, miller, dz)[0]
        nlayers = max(layers) + 1
        fig, axs = plt.subplots(nlayers, 1, figsize = (10, nlayers*2), sharex = True)
        indexs = range(len(atoms))
        images = []
        xindex = (self.pdos_energies>Emin) & (self.pdos_energies<Emax)
        for ilayer in range(nlayers):
            iax = nlayers - ilayer - 1
            # axs[iax].set_yticks([])
            index = [j for j in indexs if layers[j] ==  ilayer]
            images.append(atoms[index])
            pdos_kinds = self.merge_kinds(index)
            self.plot_pdos(energies = self.pdos_energies, pdos_kinds = pdos_kinds, select = select, 
                           Emin = Emin, Emax = Emax, ax = axs[iax], 
                           legend = False, xylabel = False, fill = fill, smearing = smearing)
            axs[iax].axvline(0, color = 'b')
            # print(iax, ilayer)
            # print(index)
            # axs[iax].set_xlim([Emin, Emax])
        plt.xlabel('E - E$_{Fermi}$ (eV)', size='16')
        plt.subplots_adjust(hspace=0)
        fig.text(0.05, 0.5, 'PDOS (a.u.)', size='16', va='center', rotation='vertical')
        if output is not None:
            # output = '{0}-pdos.png'.format(self.prefix) 
            plt.savefig('%s'%output)
        # view(images)
        return axs, images
    def get_pdos(self, species, orbital):
        dos = self.pdos_kinds[species][orbital]
        return self.pdos_energies, dos
    def get_d_band_center(sef, species = None, orbital = 'd'):
        pdos_kinds
        energies, dos = self.get_pdos(species, orbital)
        Nstates = np.trapz(dos, energies)
        occupied = energies <= 0.0
        N_occupied_states = np.trapz(dos[occupied], energies[occupied])
        # first moment
        ed = np.trapz(energies * dos, energies) / Nstates
        # second moment
        wd2 = np.trapz(energies**2 * dos, energies) / Nstates
        return d_center, d_width
    def smearing(self, energies, dos, sigma = 0.1, de=0.01, total = False):
        '''
        'natoms':         the total number of atoms of this kind in the structure
        'sigma':          sigma for the gaussian distribution (default: 0.01)
        'de':     integration step size (default: 0.01)
        '''
        npnts = len(energies)
        emin = min(energies)
        emax = max(energies)
        nmesh = int((emax-emin)/de)+1   
        xmesh = np.linspace(emin, emax, nmesh)
        ymesh = np.zeros((nmesh, 1))
        fact = de/(sigma*np.sqrt(2.0*np.pi))
        for i in range(nmesh):
            func = np.exp(-(xmesh[i]-energies)**2/(2.0*sigma**2))*fact
            ymesh[i] = func.dot(dos[:])
        # ymesh /= natoms  # normalize
        # return energies, dos
        return xmesh, ymesh
    
def compare_pdos(dos1, index1, dos2, index2):
    '''
    '''
    if len(index1) != len(index2):
        assert ('number of atoms wrong!')
    natoms = len(index1)
    for i in range(natoms):
        if dos1.atoms[i].symbol != dos2.atoms[i].symbol:
            assert('atoms symbol not match!')
    ddos = copy.deepcopy(dos1)
    # print(dos1.atoms.get_chemical_symbols())
    # print(dos2.atoms.get_chemical_symbols())
    atoms = ddos.atoms
    for i in range(natoms):
        iatom1 = index1[i]
        iatom2 = index2[i]
        for channel in ddos.pdos_atoms[iatom1 + 1]:
            # print(i, atoms[i].symbol, channel)
            npoints = len(ddos.pdos_atoms[iatom1 + 1][channel][0])
            ddos.pdos_atoms[iatom1 + 1][channel] = dos1.pdos_atoms[iatom1 + 1][channel][:, 0:npoints] - dos2.pdos_atoms[iatom2 + 1][channel][:, 0:npoints]
    ddos.pdos_kinds = ddos.merge_kinds(index = range(len(ddos.atoms)))
    return ddos


from ase import Atoms
from ase import io
from ase.build import bulk, sort
from ase.io import read
from xespresso import Espresso
from xespresso.hpxespresso import HpXEspresso
from itertools import chain
import numpy as np
import copy
import os, shutil
import sys

def update_UVscsd(U_dict,Hubbard_site):
    """
    Updates the labels of the Hubbard sites in the structure on the base of the U value.
    If there are more than 9 different Hubbard sites, letters and not numbers will be used to
    label the site becausepw.x cannot distinguish for example "Ti1" from "Ti12".
    The different number of Hubbard sites cannot exceed 26 (letters of the English alphabet)
    
    WARNING: By default pw.x can deal only up to 10 different Hubbard atoms. If you need more you need to
    change the source code and re-compile your pw.x code.
    
    U_dict: dict
        dictionary obtained with the read_Hubbard_parameters function
    Hubbard_site: str
        A str indicatint the type of the Hubbard specie
        
    Returns:
        - U: dict
            A copy of U_dict in which the new_label is modified according on the U values 
    """
    
    U = copy.deepcopy(U_dict)
    import string as STR
    alphabet = list(STR.ascii_lowercase)


    new_labels = []
    new_labels = []
    Hubbard_U = {}


    count_sites_spin0 = []
    count_sites_spin1 = []
    count_sites_spinm1 = []
    for site, info in U.items():
        if info["spin"] == "1":
            count = info["new_label"][len(Hubbard_site):]
            if count == "":
                count = 1
            else:
                count = int(count)
            count_sites_spin1.append(count)
        elif info["spin"] == "-1":
            count = info["new_label"][len(Hubbard_site):]
            if count == "":
                count = 1
            else:
                count = int(count)
            count_sites_spinm1.append(count)
        elif info["spin"] == "0":
            count = info["new_label"][len(Hubbard_site):]
            if count == "":
                count = 1
            else:
                count = int(count)
            
            count_sites_spin0.append(count)
        else:
            pass
            

    count_sites_spin0 = len(count_sites_spin0) 
    count_sites_spin1 = len(count_sites_spin1)
    count_sites_spinm1 = len(count_sites_spinm1)


    new_labels = []
    for site, info in U.items():
        if info["spin"] == "1":
            count = info["new_label"][len(Hubbard_site):]
            if count == "":
                count = 1
            else:
                count = int(count)
            new_labels.append(count)
            info["new_new_labels"] = count
 
        elif info["spin"] == "-1":
            count = info["new_label"][len(Hubbard_site):]
            if count == "":
                count = 1
            else:
                count = int(count)
            new_count = count + count_sites_spin1
            new_labels.append(new_count)
            info["new_new_labels"] = new_count

        else:
            count = info["new_label"][len(Hubbard_site):]
            if count == "":
                count = 1
            else:
                count = int(count)
            new_count = count + count_sites_spin1  + count_sites_spinm1
            new_labels.append(new_count)
            info["new_new_labels"] = new_count

            
    new_labels.sort()
    new_labels =set(new_labels)
    indeces = {}
    
    if len(new_labels) >  len(alphabet):
        sys.exit("I can differentiate maximum 26 sites")
        
    for n,i in enumerate(new_labels):
        indeces[i] = n+1
    #print(indeces)
    
    
    for site, info in U.items():
        count = info["new_new_labels"]
        indx = indeces[count]
        #print(count,indx)
        if indx < 10:
            info["new_label"] = Hubbard_site + str(indx)
        else:
            info["new_label"] = Hubbard_site + alphabet[indx]
            
        
    return U 

def update_Hubbard_parameters(U_dict,pseudopotentials,Hubbard_site,atoms): 
    """
    Updating the input_ntyp, pseudopotentials, and structure on the base of the new_labels/U values
    
    U_dict: dict
        dictionary obtained with the read_Hubbard_parameters function
        
    pseudopotentials: dict
        A dict specifing the pseudopotential name per atom type
    
    Hubbard_site: str
        A str indicatint the type of the Hubbard specie
        
    atoms: ase.atoms
        An ase structure object
        
    Returns:
        - dict with the updated structure, pseudopotentials, and input_ntyp dictionary  
    """
    Hubbard_U = dict()
    starting_magnetization = dict()
    for site, info in U_dict.items():
        label = info["new_label"]
        if label not in Hubbard_U:
            Hubbard_U[label] = float(info["U"])
            starting_magnetization[label] = float(info["spin"])
            
    remove_species = []
    for elem, pseudo in pseudopotentials.items():
        if Hubbard_site in elem:
            pseudo_name = pseudopotentials[elem]
            remove_species.append(elem)
            
    for elem in remove_species:    
        del pseudopotentials[elem]
            
    for site in Hubbard_U.keys():
        pseudopotentials[site] = pseudo_name
        
    atoms.info['species'] = atoms.get_chemical_symbols()
    
    for site, info in U_dict.items():
        atoms.info['species'][int(site)-1] = info["new_label"]
    
    return {"Hubbard_U": Hubbard_U,
            "starting_magnetization":starting_magnetization,
            "pseudopotentials"  : pseudopotentials,
           "atoms": atoms}

class Uscsd():
    def __init__(self, restart=None, ignore_bad_restart_file=False,
                 label='xespresso', atoms=None, package = 'pw', parallel =  '',
                 queue = None,queue_hp = False, parallel_hp = '',max_iter= 1, iteration=1, sc_type =None,
                 insulator = None, magnetic = None, nbnd =  None, relax_type = "vc-relax",
                 parallelization_hp = False, step = None, Hubbard_site = None, conv_Uscsd = 0.01,
                 ethr_relax = None, ethr_scf = None, ethr_scf2 = None,
                 **kwargs):
        """
        This class allows to automate the self-consistent (site-dependent) calculation of U/V values
        Automating the necessary relax-scf-hp procedure. In particular, one iteration requires:
        
            - For a metal:
                1) relaxation (optional)
                2) scf calculation with "smaering"(on the optimized structure if step 1) is performed)
                3) hp calculation
                
            - For a non-magnetic insulator:
                1) relaxation (optional)
                2) scf calculation with "fixed"(on the optimized structure if step 1) is performed)
                3) hp calculation
                
            - For a magnetic insulator:
                1) relaxation (optional)
                2) scf calculation with "smaering"(on the optimized structure if step 1) is performed)
                3) scf calculation with "fixed" reading the total_magnetization and the number of bands 
                   from step 1
                4) hp calculation
                
        When the U/V values are obtained, the subsequent iteration is run using the new U/V values and eventually 
        the structure optimized in step 1) until a maximum numer of iterations (max_iter) is performed or until
        U/V convergence is reached.
        
        The class is able to detect the procedure to apply. WARNING: When sc_type = "scf" and iteration = 1,
        you need to provide values for "insulator", "magnetic", "nbnd".
        
        The class is able to restart an interrupted self-consistent calculation of U/V if for one iteration
        only the "relax" or "scf" step perfomed. 
        WARNING: restart from an incomplete "relax" step is not yet possible.
        
        
        Check:
        - "Hubbard parameters from density-functional perturbation theory", 
        Phys. Rev. B 98, 085127 (2018); arXiv:1805.01805 using ase and XEspresso, 
        of which pXEspresso is a subclass.
        - C. Ricca, I. Timrov, M. Cococcioni, N. Marzari, and U. Aschauer,
          "Self-consistent site-dependent DFT+U study of stoichiometric and defective SrMnO3",
          Phys. Rev. B 99, 094102 (2019); arXiv:1811.10858
        - C. Ricca, I. Timrov, M. Cococcioni, N. Marzari, and U. Aschauer,
          "Self-consistent DFT+U+V study of oxygen vacancies in SrTiO3",
          Phys. Rev. Research 2, 023313 (2020); arXiv:2004.04142
        
        input_data, pseudopotentials, kspacing, kpts, koffset
            Please have a look at Espresso module in ASE
            
        max_iter: int
            An integer indicating the maximum number of iteration
            
        Hubbard_site: str
            A str indicating the atomic species of the Hubbard atom for which the self-consistent (site-dependent)
            procedure is to be performd. Example: Hubbard_site = "Ti".
            
        conv_Uscsd: float
            A float indicating he threshold to be used to verify convergence of the U/V values
        
        !!!!!!!!!!!!!!!!! WARNING !!!!!!!!!!!!!!!!!!!
        Do not forget to provide in the input_data dictionry and two elements:
            "lda_plus_U" : True
            "input_ntyp": { "starting_magnetization" : {Hubbard_site: XX},
                            "Hubbard_U" : {Hubbard_site: XX}}
        """
        
        self.label_ref = label

        self.directory = os.path.split(label)[0]
        self.prefix = os.path.split(label)[1]
        self.directory_ref = self.directory
        
        self.queue_hp = queue_hp
        self.parallel_hp = parallel_hp
        self.queue = queue
        self.parallel = parallel
            
        self.step = None
        self.iteration = iteration
        self.max_iter = max_iter
        
        if sc_type == None:
            self.sc_type = "relax"
        else:
            self.sc_type = sc_type
        

        self.relax_type = relax_type

        if self.sc_type != None and self.sc_type not in ["relax", "scf"]:
            sys.exit("Please provide a str for the sc_type variable: relax or scf")

        if self.relax_type != None and self.relax_type not in ["relax", "vc-relax"]:
            sys.exit("Please provide a str for the relax_type variable: relax or vc-relax")
        if self.sc_type != None and self.sc_type not in ["relax", "scf"]:
            sys.exit("Please provide a bool for the sc_type variable: relax or scf")

        self.insulator = insulator
        self.magnetic = magnetic
        self.nbnd = nbnd

        if self.insulator != None and bool(self.insulator) == False:
            sys.exit("Please provide a bool for the insulator variable")

        if self.magnetic != None and self.magnetic not in [1,2]:
            sys.exit("Please provide an int for the magnetic  variable: 1 for spin unpolirzed and 2 for spin poolariz")

        if self.nbnd != None and type(self.nbnd) != int:
            sys.exit("Please provide an int for the nbnd variable")


        self.parallelization_hp = parallelization_hp
    
        
        self.converged = False
        
        self.kwargs = kwargs
        
        self.atoms = atoms
        
        self.UVscsd = {}
        self.etot = {}
        self.opt_structures = {}
        
        if Hubbard_site == None:
            sys.exit("You need to specify the Hubbard_site ")
        else:
            self.Hubbard_site = Hubbard_site
            
        self.conv_Uscsd = conv_Uscsd
        
        self.ethr_relax = ethr_relax
        self.ethr_scf = ethr_scf
        self.ethr_scf2 = ethr_scf2
        
        
    def converged_U(self):
        old_U = []
        new_U = []

        for site, info in self.UVscsd[self.iteration-1].items():
            old_U.append(float(info["U"]))

        for site, info in self.UVscsd[self.iteration].items():
            new_U.append(float(info["U"]))

        #comparison = np.isclose(np.array(old_U),np.array(new_U),conv_Uscsd)
        self.converged = np.allclose(np.array(old_U),np.array(new_U),self.conv_Uscsd)       
        
    def check_iteration(self):
        """
        In case previous relax-scf-hp iterations have been previously performed, this function
        identifies the last iteration and which is the last step at which that iteration was stopped
        
        WARNING:
        - Not yet able to identify if structural relaxation did not convrge and needs restart
        
        TO DO:
            - make it smarter being able to recognize only "relax_*", "scf_*", "hp_*"
        
        """
        
        folders = [name for name in os.listdir(self.directory) if os.path.isdir(os.path.join(self.directory,name))]
        for folder_path in folders:
            if len(os.listdir(os.path.join(self.directory,folder_path))) == 0: # Check is empty..
                shutil.rmtree(os.path.join(self.directory,folder_path)) # Delete..
                folders.remove(folder_path)
  
        Folders = []
        for folder_path in folders:
            if "_"  in folder_path:
                Folders.append(folder_path)
            else:
                pass
                #folders.remove(folder_path
        
        if not folders:
            dirs = []
        else:
            dirs =  [int(name.split('_')[1]) for name in Folders]
 
        if not dirs:
            self.iteration = 1
            self.step = "re-start"
        else:
            self.iteration = max(dirs)

            if os.path.exists(os.path.join(self.directory,"hp_"+str(self.iteration))):
                self.step = "hp"
            elif os.path.exists(os.path.join(self.directory,"scf_"+str(self.iteration))):
                self.step = "scf"
                print("""WARNING: For magnetic insulators, restart of the second scf step with 'fixed'
                from the previous scf step with 'smearing' is not yet implemented. In this case, the
                hp step will fail if you do not remove also the scf_$iteration folder.""")
            elif os.path.exists(os.path.join(self.directory,"relax_"+str(self.iteration))):
                self.step = "relax"
            else:
                self.step = "re-start"
                

            
    def run_sc(self):
        """
        Main function wunning the self-consistence procedure
        """
        
        if os.path.exists(self.directory) == False:
            os.mkdir(self.directory)
            print("Creating working directory")
            
        
        self.check_iteration()
        
        print("-------------------------------------------")            
        print("Initial Iteration:",self.iteration,self.step)
        print("-------------------------------------------")
        
        calc = HpXEspresso(label=self.label_ref,
                           parallel = self.parallel,
                           parallel_hp = self.parallel_hp,
                           queue = self.queue,
                           queue_hp = self.queue_hp,
                           max_iter = self.max_iter,
                           iteration = self.iteration,
                           sc_type = self.sc_type,
                           relax_type = self.relax_type,
                           insulator = self.insulator,
                           magnetic = self.magnetic,
                           nbnd = self.nbnd,
                           parallelization_hp = self.parallelization_hp,
                           step = self.step,
                           ethr_relax = self.ethr_relax,
                           ethr_scf = self.ethr_scf,
                           ethr_scf2 = self.ethr_scf2,
                           **self.kwargs)
        
        if self.iteration > 1:
            U = calc.read_Hubbard_parameters(self.iteration-1)
            self.UVscsd[self.iteration-1] =  U
            
            if self.step == 'hp':
                U = calc.read_Hubbard_parameters(self.iteration)
                self.UVscsd[self.iteration] =  U
                self.converged_U()
                print("U/V values converged: {}".format(self.converged))
                if self.converged == True:
                    if self.sc_type == "relax":
                        directory = os.path.join(self.directory_ref,"relax_"+str(self.iteration))
                    else:
                        directory = os.path.join(self.directory_ref,"scf_"+str(self.iteration))
                    calc.read_results_fromdirectory(directory)
                    self.etot[self.iteration] = calc.get_potential_energy()
                    self.opt_structures[self.iteration] = calc.atoms
                    sys.exit("Stopping: Convergence achieved!")
            
            new_U = update_UVscsd(U,self.Hubbard_site)
            new_data = update_Hubbard_parameters(new_U,self.kwargs['pseudopotentials'],self.Hubbard_site,self.atoms)
            self.atoms = new_data["atoms"]
            self.kwargs['pseudopotentials'] = new_data['pseudopotentials']
            self.kwargs['input_data']['input_ntyp']['Hubbard_U'] = new_data['Hubbard_U']
            self.kwargs['input_data']['input_ntyp']['starting_magnetization'] = new_data['starting_magnetization']
        else:
            pass

        atoms = self.atoms

            
        if self.step == "re-start":
            if self.sc_type == "relax":
                print("- Running relaxation step")
                calc.run_relax()
                atoms.calc = calc
                self.atoms.set_calculator(calc)
                e_relax = self.atoms.get_potential_energy()
                calc.read_results()
                self.etot[self.iteration] = e_relax
                self.opt_structures[self.iteration] = calc.atoms
            
            else:
                pass
           
            print("- Running (first) scf step")
            calc.run_scf()
            atoms.calc = calc
            atoms.set_calculator(calc)
            e_scf = atoms.get_potential_energy()
            calc.read_results()
            if self.sc_type == "scf":
                self.etot[self.iteration] = e_scf
                self.opt_structures[self.iteration] = calc.get_potential_energy()
            
            print("- Running second scf step")
            calc.run_scf2()
            if calc.scf2_necessary == True:
                atoms.set_calculator(calc)
                atoms.get_potential_energy()
            else:
                pass
            print("- Running hp step")
            calc.read_results()
            calc.run_hp()
            
            atoms_iter = calc.atoms

        elif self.step == "relax":
            print("- Restarting from (first) scf step")
            #calc.clean()
            calc.run_scf()
            atoms.calc = calc
            atoms.set_calculator(calc)
            atoms.get_potential_energy()
            calc.read_results()
                                
            self.etot[self.iteration] = calc.get_potential_energy()
            self.opt_structures[self.iteration] = calc.atoms

            
    

            print("- Running second scf step")
            calc.run_scf2()
            if calc.scf2_necessary == True:
                atoms.set_calculator(calc)
                atoms.get_potential_energy()
            else:
                pass           
 
            print("- Running hp step")
            calc.read_results()
            calc.run_hp()
            
            atoms_iter = calc.atoms

        elif self.step == "scf":
            if self.iteration ==1:
                pass
                calc.read_results()
            else:
                if self.sc_type == "relax":
                    old_directory = os.path.join(self.directory_ref,"relax_"+str(self.iteration))
                else:
                    old_directory = os.path.join(self.directory_ref,"scf_"+str(self.iteration))
                calc.read_results_fromdirectory(old_directory)
                self.etot[self.iteration] = calc.get_potential_energy()
                self.opt_structures[self.iteration] = calc.atoms

            
            print("- Restarting from hp step")
            calc.run_hp()
            atoms_iter = calc.atoms

        elif self.step == "hp":
            if self.sc_type == "relax":
                old_directory = os.path.join(self.directory_ref,"relax_"+str(self.iteration))
            else:
                old_directory = os.path.join(self.directory_ref,"scf_"+str(self.iteration))
            calc.read_results_fromdirectory(old_directory)
            self.etot[self.iteration] = calc.get_potential_energy()
            self.opt_structures[self.iteration] = calc.atoms
            atoms_iter = calc.atoms
                
           
        print("- Reading Hubbard U/V values")
        U = calc.read_Hubbard_parameters(self.iteration)
        self.UVscsd[self.iteration] =  U
        
        if self.iteration > 1:
            self.converged_U()
        else:
            pass
        print("U/V values converged: {}".format(self.converged))
        
        
        self.iteration += 1
        self.step = "re-start"
        
        new_U = update_UVscsd(U,self.Hubbard_site)
        new_data = update_Hubbard_parameters(new_U,self.kwargs['pseudopotentials'],self.Hubbard_site,atoms_iter)
        atoms_iter = new_data["atoms"]
        self.kwargs['pseudopotentials'] = new_data['pseudopotentials']
        self.kwargs['input_data']['input_ntyp']['Hubbard_U'] = new_data['Hubbard_U']
        self.kwargs['input_data']['input_ntyp']['starting_magnetization'] = new_data['starting_magnetization']
        

        
        while self.converged == False and self.iteration <= self.max_iter:
            print("-------------------------------------------")
            print("Iteration:",self.iteration,self.step)
            print("-------------------------------------------")
            calc = HpXEspresso(label=self.label_ref,
                           parallel = self.parallel,
                           parallel_hp = self.parallel_hp,
                           queue = self.queue,
                           queue_hp = self.queue_hp,
                           max_iter = self.max_iter,
                           iteration = self.iteration,
                           sc_type = self.sc_type,
                           relax_type = self.relax_type,
                           insulator = self.insulator,
                           magnetic = self.magnetic,
                           nbnd = self.nbnd,
                           parallelization_hp = self.parallelization_hp,
                           step = self.step,
                          ethr_relax = self.ethr_relax,
                           ethr_scf = self.ethr_scf,
                           ethr_scf2 = self.ethr_scf2,
                               **self.kwargs)
            

            if self.sc_type == "relax":
                print("- Running relaxation step")
                calc.run_relax()
                atoms_iter.calc = calc
                atoms_iter.set_calculator(calc)
                e_relax = atoms_iter.get_potential_energy()
                calc.read_results()
                self.etot[self.iteration] = e_relax
                self.opt_structures[self.iteration] = calc.atoms
            else:
                atoms_iter = self.atoms
                atoms_iter.calc = calc 
                atoms_iter.info['species'] = atoms_iter.get_chemical_symbols()
                for site, info in new_U.items():
                    atoms_iter.info['species'][int(site)-1] = info["new_label"]


            print("- Running (first) scf step")    
            calc.run_scf()
            atoms_iter.set_calculator(calc)
            e_scf = atoms_iter.get_potential_energy()
            calc.read_results()
            if self.sc_type == "scf":
                self.etot[self.iteration] = e_scf
            
            print("- Running second step")
            calc.run_scf2()
            atoms_iter.set_calculator(calc)
            if calc.scf2_necessary == True:
                atoms_iter.set_calculator(calc)
                atoms_iter.get_potential_energy()
            else:
                pass            

            print("- Running hp step")
            calc.read_results()
            calc.run_hp()
            
            print("- Reading Hubbard U/V values")
            U = calc.read_Hubbard_parameters(self.iteration)
            self.UVscsd[self.iteration] = U

            if self.sc_type == "relax":
                atoms_iter = calc.atoms
                
            self.converged_U()
            print("U/V values converged: {}".format(self.converged))
                
            self.iteration += 1
            self.step = "re-start"
            

            new_U = update_UVscsd(U,self.Hubbard_site)
            new_data = update_Hubbard_parameters(new_U,
                                                 self.kwargs['pseudopotentials'],
                                                 self.Hubbard_site,
                                                 atoms_iter)
            atoms_iter = new_data["atoms"]
            self.kwargs['pseudopotentials'] = new_data['pseudopotentials']
            self.kwargs['input_data']['input_ntyp']['Hubbard_U'] = new_data['Hubbard_U']
            self.kwargs['input_data']['input_ntyp']['starting_magnetization'] = new_data['starting_magnetization']
            




import numpy as np
from xespresso import Espresso
import matplotlib.pyplot as plt
import os
import shutil
import subprocess

class COHP:
    def __init__(self, directory = '.', prefix = 'cohp', indexs = [[1, 2]], queue = False, command = 'lobster', **kwargs):
        """
        Energetic window [COHPstartEnergy, COHPendEnergy]
        """
        self.directory = directory
        self.prefix = prefix
        self.parameters = kwargs
        self.indexs = indexs
        self.command = command
    def run(self, cpu = 1):
        '''
        '''
        shutil.copy(os.path.join(self.directory, '%s.pwi'%self.prefix), os.path.join(self.directory, '%s.scf.in'%self.prefix))
        self.write_input()
        command = self.command
        os.system('export OMP_NUM_THREADS=%s'%cpu)
        print('Running %s on %s cpus' %(command, cpu))
        try:
            proc = subprocess.Popen(command, shell=True, cwd=self.directory)
        except OSError as err:
            msg = 'Failed to execute "{}"'.format(command)
            raise EnvironmentError(msg) from err

        errorcode = proc.wait()

        if errorcode:
            path = os.path.abspath(self.directory)
            msg = ('Command "{}" failed in '
                   '{} with error code {}'.format(command,
                                                  path, errorcode))
            print(msg)
            exit()
        print('Done: %s' %command)
        #
    def write_input(self, ):
        filename = os.path.join(self.directory, 'lobsterin')
        with open(filename, 'w') as f:
            for key, value in self.parameters.items():
                if isinstance(value, list):
                    for subvalue in value:
                        f.write('  %s  %s \n' %(key, subvalue))
                else:
                    f.write('  %s  %s \n' %(key, value))
            for ind in self.indexs:
                f.write('cohpbetween atom %s atom %s \n' %(ind[0], ind[1]))
    def read_cohp(self, ):
        from ase.calculators.vasp import VaspDos
        dos = VaspDos(os.path.join(self.directory, 'DOSCAR.lobster'))
        self.dos = dos._total_dos[1]
        self.dos_energies = dos._total_dos[0]
        #
        datas = np.genfromtxt(os.path.join(self.directory, 'COHPCAR.lobster'), skip_header = 3 + len(self.indexs))
        self.cohp = datas[:,1]
        self.cohp_energies = datas[:,0]
        #
        datas = np.genfromtxt(os.path.join(self.directory, 'COOPCAR.lobster'), skip_header = 3 + len(self.indexs))
        self.coop = datas[:,1]
        self.coop_energies = datas[:,0]
    def read_icohp(self, ):
        from ase.calculators.vasp import VaspDos
        datas = np.genfromtxt(os.path.join(self.directory, 'ICOOPLIST.lobster'), skip_header = 1 + len(self.indexs))
        self.icohp = datas
        # self.icohp_energies = datas[:, 7]
        #
        datas = np.genfromtxt(os.path.join(self.directory, 'ICOOPLIST.lobster'), skip_header = 1 + len(self.indexs))
        self.icoop = datas
        # self.icoop_energies = datas[:, 7]
    def plot_cohp(self, ax = None, output = None):
        '''
        '''
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots(1, 2)
        ax[0].plot(self.dos, self.dos_energies)
        ax[0].set_ylabel('Energy (eV)')
        ax[1].plot(-self.cohp, self.cohp_energies)
        ax[1].axvline(x = 0, color = 'k')
        ax[0].set_xlabel('DOS')
        ax[1].set_xlabel('-COHP')
        if output:
            plt.savefig(output)
        return ax
    def plot_coop(self, ax = None, output = None):
        '''
        '''
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots(1, 2)
        ax[0].plot(self.dos, self.dos_energies)
        ax[0].set_ylabel('Energy (eV)')
        ax[1].plot(self.coop, self.coop_energies)
        ax[1].axvline(x = 0, color = 'k')
        ax[0].set_xlabel('DOS')
        ax[1].set_xlabel('COOP')
        if output:
            plt.savefig(output)
        return ax
    
    


"""Quantum ESPRESSO Calculator

export PYTHONPATH="Your-Location"/xespresso:$PYTHONPATH
export ASE_ESPRESSO_COMMAND="/path/to/PACKAGE.x  PARALLEL  -in  PREFIX.PACKAGEi  >  PREFIX.PACKAGEo"
export ESPRESSO_PSEUDO="/path/to/pseudo"

Run PACKAGE.x jobs.
"""


from ase import constraints, io
from ase.calculators.calculator import FileIOCalculator, CalculationFailed, equal, compare_atoms
from ase.io.formats import F
from numpy.core.fromnumeric import sort
import pseudo
from xespresso.xio import write_espresso_in, read_espresso_input, read_espresso_asei, write_espresso_asei, get_atomic_species, get_atomic_constraints
import os
import shutil
import numpy as np
from datetime import datetime
import copy
import logging
import sys
import subprocess
import warnings
from pprint import pprint

logging.basicConfig(stream=sys.stdout,
                    format=('%(levelname)-8s '
                            '[%(funcName)-20s]: %(message)s'),
                    level=logging.INFO)

logger = logging.getLogger('xespresso')

config_files = [os.path.join(os.environ['HOME'], '.xespressorc'),
            '.xespressorc']
error_template = 'Property "%s" not available. Please try running Quantum\n' \
                 'Espresso first by calling Atoms.get_potential_energy().'

warn_template = 'Property "%s" is None. Typically, this is because the ' \
                'required information has not been printed by Quantum ' \
                'Espresso at a "low" verbosity level (the default). ' \
                'Please try running Quantum Espresso with "high" verbosity.'



class Espresso(FileIOCalculator):
    implemented_properties = ['energy', 'forces', 'stress', 'magmoms', 'time']
    command = 'PACKAGE.x  PARALLEL  -in  PREFIX.PACKAGEi  >  PREFIX.PACKAGEo'
    discard_results_on_any_change = False

    def __init__(self, label='xespresso', prefix = None, atoms=None, package = 'pw', parallel = '',
                 queue = None, debug = False, **kwargs):
        """
        Accepts all the options for pw.x as given in the QE docs, plus some
        additional options:

        input_data, pseudopotentials, kspacing, kpts, koffset
            Please have a look at Espresso module in ASE
        queue: dict
            A dictionary with parameters for job submission, e.g.
             queue = {'nodes': 4, 'ntasks-per-node': 20, 
                      'account': 'xxx', 'partition': 'normal', 
                      'time': '23:59:00'}
        package: str
            Choose the quantum espresso package: pw, dos, projwfc, band, pp, ph, ..
            For NEB calculation, please use neb.NEBEspresso module.
            Calculaiton use phonon is not implemented yet.
        parallel: str
            A str which control the parallelization parameters: -nimage, -npools, 
            -nband, -ntg, -ndiag or -northo (shorthands, respectively: 
            -ni, -nk, -nb, -nt, -nd).

        General working procedure is as follows:
            1. Perform a regular self-consistent calculation:
              >>> calc = Espresso(input_data=input_data, ...)
              >>> atoms.set_calculator(calc)
              >>> atoms.get_potential_energy()
            2. post calculation can be made as follows:
              >>> calc.post('package' = 'dos', **kwargs)
              >>> calc.post('package' = 'pp', **kwargs)
            3. non-self-consistent calculation can be made as follows:
              >>> calc.nscf(calculation = 'nscf', kpts=(4, 4, 4))
              >>> calc.nscf_calculate()
        """
        print("{0:=^60}".format(package))
        self.logger = logger
        if debug:
            self.logger.setLevel(debug)
        self.set_label(label, prefix)
        self.scf_directory = None
        self.scf_parameters = None
        self.scf_results = None
        kwargs = self.check_input(kwargs)
        FileIOCalculator.__init__(self, restart = self.directory, 
                                  label = self.label, atoms = atoms, **kwargs)
        if atoms:
            self.atoms = atoms
        self.queue = queue
        self.parallel = parallel
        self.package = package
        self.parallel = parallel
        self.debug = debug
        # self.discard_results_on_any_change = False
        
    def set_label(self, label, prefix):
        '''
        set directory and prefix from label
        '''
        self.directory = label
        if not prefix:
            self.prefix = os.path.split(label)[1]
        else:
            self.prefix = prefix
        if not os.path.exists(self.directory):
            os.makedirs(self.directory)
        self.label = os.path.join(self.directory, self.prefix)
        self.pwi = os.path.join(self.directory, '%s.pwi'%self.prefix)
        self.pwo = os.path.join(self.directory, '%s.pwo'%self.prefix)
        self.asei = os.path.join(self.directory, '%s.asei'%self.prefix)
        self.asei_temp = os.path.join(self.directory, '.%s.asei_temp'%self.prefix)
        self.save_directory = os.path.join(self.directory, '%s.save'%self.prefix)
        self.logger.debug('Directory: %s'%(self.directory))
        self.logger.debug('Prefix: %s'%(self.prefix))
    def check_pseudopotentials(self, pseudopotentials):
        """
        """
        # pseudopotentials
        if 'species' not in self.atoms.info:
            all_species = set(self.atoms.get_chemical_symbols())
        else:
            all_species = set(self.atoms.info['species'])
        new_pseudopotentials = {}
        for species in all_species:
            assert species in pseudopotentials, '\n  Species: %s, does not have a pseudopotentials!'%(species)
            new_pseudopotentials[species] = pseudopotentials[species]
    def check_input(self, kwargs, package = 'PW'):
        """
        Chekc input paramter type.
        set parameters for pw.x
        all other parameters are stored in 'input_data'
        """
        from xespresso.xio import sort_qe_input, check_qe_input
        sorted_parameters, unuse_parameters = sort_qe_input(kwargs)
        check_qe_input(sorted_parameters['input_data'])
        if unuse_parameters:
            self.logger.debug('Warnning: parameter %s not used, please check.'%unuse_parameters)
        sorted_parameters['input_data']['CONTROL']['prefix'] = self.prefix
        sorted_parameters['input_data']['CONTROL']['verbosity'] = 'high'
        ase_parameters = sorted_parameters
        self.ase_parameters = ase_parameters
        # pprint(sorted_parameters)
        return sorted_parameters
    def set_queue(self, package = None, parallel = None, queue = None, command = None):
        '''
        If queue, change command to "sbatch .job_file".
        The queue information are written into '.job_file'
        '''
        if queue is None:
            queue = self.queue
        else:
            self.queue = queue
        if package is None:
            package = self.package
        else:
            self.package = package
        if parallel is None:
            parallel = self.parallel
        else:
            self.parallel = parallel
        if command is None:
            command = os.environ.get('ASE_ESPRESSO_COMMAND')
        if 'PACKAGE' in command:
            if 'pw' in package:
                command = command.replace('PACKAGE', package, 1)
                command = command.replace('PACKAGE', 'pw', 2)
            else:
                command = command.replace('PACKAGE', package)
        if 'PREFIX' in command:
            command = command.replace('PREFIX', self.prefix)
        if 'PARALLEL' in command:
            command = command.replace('PARALLEL', parallel)
        self.logger.debug('Espresso command: %s'%(command))
        if queue:
            script = ''
            if 'config' in queue:
                cf = os.path.join(os.environ['HOME'], queue['config'])
                file = open(cf, 'r')
                script = file.read()
                # del queue['config']
            else:
                for cf in config_files:
                    if os.path.exists(cf):
                        file = open(cf, 'r')
                        script = file.read()
            jobname = self.prefix
            with open('%s/.job_file' % self.directory, 'w') as fh:
                fh.writelines("#!/bin/bash\n")
                fh.writelines("#SBATCH --job-name=%s \n" % jobname)
                fh.writelines("#SBATCH --output=%s.out\n" % self.prefix)
                fh.writelines("#SBATCH --error=%s.err\n" % self.prefix)
                fh.writelines("#SBATCH --wait\n")
                for key, value in queue.items():
                    if key == 'config': continue
                    if value:
                        fh.writelines("#SBATCH --%s=%s\n" %(key, value))
                fh.writelines("%s \n"%script)
                fh.writelines("%s \n" % command)
            self.command = "sbatch {0}".format('.job_file')
        else:
            self.command = command
        self.logger.debug('Queue command: %s'%(self.command))
    def read(self, label):
        """Read the files in a calculation if they exist.
        This function reads self.parameters and atoms.
        """
        # Else read a regular calculation. we start with reading stuff
        # that is independent of the calculation state.
        self.directory = label
        self.restart_atoms = None
        self.restart_parameters = {}
        if os.path.exists(label) and os.path.exists(self.asei):
            self.logger.debug('Try to read information from folder: %s.'%(label))
            try:
                # atoms, parameters = self.read_xml_file()
                atoms, parameters = read_espresso_asei(self.asei)
                self.restart_atoms = atoms
                self.restart_parameters = copy.deepcopy(parameters)
                # self.parameters = copy.deepcopy(parameters)
                self.read_results()
            except Exception as e:
                self.logger.debug('Directory: %s exist, faild to read. %s'%(label, e))
        else:
            self.logger.debug('Directory %s is not a espresso folder, start a new calculation:'%(label))
    def set(self, **kwargs):
        """
        """
        from xespresso.input_parameters import restart_ignore
        self.changed_parameters, self.igonre_parameters = self.compare_parameters(self.restart_parameters, self.ase_parameters, ignore=restart_ignore['PW'])
        self.parameters = kwargs
        if self.discard_results_on_any_change and self.changed_parameters:
            self.reset()
        return self.changed_parameters
        
    def check_state(self, atoms, tol=1e-4):
        """Check for any system changes since last calculation."""
        if not os.path.exists(self.asei):
            return True
        if isinstance(self.restart_atoms, list):
            if len(self.restart_atoms) != len(atoms):
                return True
            for i in range(len(atoms)):
                if compare_atoms(self.restart_atoms[i], atoms[i], tol=tol,
                             excluded_properties=set(self.ignored_changes)):
                    self.logger.debug('Atoms changed, run a new calculation')
                    return True
        else:
            if compare_atoms(self.restart_atoms, atoms, tol=tol,
                             excluded_properties=set(self.ignored_changes)):
                self.logger.debug('Atoms changed, run a new calculation')
                return True
        # if self.igonre_parameters:
            # self.logger.debug('Parameter: %s are ignored, results are not affected.'%self.igonre_parameters)
        if self.changed_parameters:
            self.logger.debug('Parameter: %s changed, results are affected'%self.changed_parameters)
            return True
        self.logger.debug('Same geometry and parameters, use previous results.')
        return False
    def compare_parameters(self, para1, para2, ignore = []):
        """
        """
        from xespresso.input_parameters import qe_namespace, default_parameters, restart_ignore
        
        changed_parameters = []
        igonre_parameters = []
        if not para1:
            changed_parameters = ['all']
            return changed_parameters, igonre_parameters
        default_parameters = default_parameters['PW']
        # pseudopotentials
        key = 'pseudopotentials'
        try:
            for species, value in para1[key].items():
                if value != para2[key][species]:
                    changed_parameters.append(key)
                    continue
        except:
            changed_parameters.append(key)
        # kpts
        key = 'kpts'
        try:
            if para1[key] != para2[key]:
                changed_parameters.append(key)
        except:
            changed_parameters.append(key)
        # input_data
        for section, paras in para1['input_data'].items():
            if section == 'INPUT_NTYP':
                changed_parameters1, igonre_parameters1 = self.compare_dict(para1['input_data'][section], para2['input_data'][section], restart_ignore['PW'])
            else:
                changed_parameters1, igonre_parameters1 = self.compare_dict(para1['input_data'][section], para2['input_data'][section], restart_ignore['PW'], default=default_parameters[section])
            changed_parameters.extend(changed_parameters1)
            igonre_parameters.extend(igonre_parameters1)
        return changed_parameters, igonre_parameters
    def compare_dict(self, dict1, dict2, ignore = [], default = None):
        igonre_parameters = []
        changed_parameters = []
        keys = set(list(dict1.keys()) + list(dict2.keys()))
        print
        for key in keys:
            if key in ignore:
                igonre_parameters.append(key)
            elif key not in dict1:
                changed_parameters.append(key)
            elif key not in dict2:
                if default and self.compare_value(dict1[key], default[key]):
                    continue
                changed_parameters.append(key)
            elif not self.compare_value(dict1[key], dict2[key]):
                changed_parameters.append(key)
        return changed_parameters, igonre_parameters
    def compare_value(self, v1, v2, tol = 1e-5):
        """
        """
        if isinstance(v1, str):
            if v1.upper() != v2.upper():
                return False
        elif isinstance(v1, dict):
            changed_parameters, igonre_parameters = self.compare_dict(v1, v2)
            if changed_parameters:
                return False
            for key, value in v1.items():
                if not self.compare_value(value, v2[key]):
                    return False
        elif isinstance(v1, bool):
            if v1 != v2:
                return False
        else:
            if abs(v1 - v2) > tol:
                return False
        return True
    def read_results(self):
        '''
        get atomic species
        '''
        pwo = os.path.join(self.directory, '%s.pwo'%self.prefix)
        convergence, meg = self.read_convergence()
        if convergence == 0:
            try:
                output = io.read(pwo)
                atomic_species = get_atomic_species(pwo)
                constraints = get_atomic_constraints(pwo, len(output))
                output.set_constraint(None)
                output.set_constraint(constraints)
                if atomic_species: output.info['species'] = atomic_species
                self.calc = output.calc
                self.results = output.calc.results
                self.results['atoms'] = output
                self.efermi = self.get_fermi_level()
                # self.nspins = self.get_number_of_spins()
                self.logger.debug('Read result successfully!')
            except Exception as e:
                self.logger.debug('Read output: %s, failed! %s'%(pwo, e))
        else:
            self.logger.debug('Not converged. %s'%(meg))
            # self.logger.debug('Read result failed!')
        # pwos = [file for file in os.listdir(self.directory) if pwo in file]
        # output = None
        # for pwo in pwos:
            # atomic_species = None
            # pwo = os.path.join(self.directory, pwo)
            # atomic_species = get_atomic_species(pwo)
            
    def write_input(self, atoms, properties=None, system_changes=None):
        FileIOCalculator.write_input(self, atoms, properties, system_changes)
        self.set_queue()
        write_espresso_in(self.label + '.pwi', atoms, **self.parameters)
        write_espresso_asei(self.label + '.asei', atoms, self.parameters)
        self.logger.debug('Write input successfully')
    def check_xml_file(self):
        '''
        If data-file-schema.xml is readable, then xml_end = True
        '''
        self.save_directory = os.path.join(self.directory, '%s.save'%self.prefix)
        if not os.path.exists(self.save_directory): return False
        xml0 = os.path.join(self.save_directory, 'data-file-schema.xml')
        xmls = [file for file in os.listdir(self.save_directory) if 'data-file-schema.xml' in file]
        goodxml = []
        for xml in xmls:
            xml = os.path.join(self.save_directory, xml)
            with open(xml, 'r') as f:
                lines = f.readlines()
                if len(lines) < 1: continue
                if '</qes:espresso>' in lines[-1]:
                    goodxml.append(xml)
        xml_end = False
        nxml = len(goodxml)
        # print(nxml, goodxml, xml0)
        if nxml > 0:
            if xml0 not in goodxml:
                shutil.copy(goodxml[-1], xml0)
            xml_end = True
        return xml_end
    def read_xml_file(self):
        from xespresso.xml_parser import xml_parser
        if not self.check_xml_file():
            self.logger.debug('xml file not finished.')
        #
        xmlfile = os.path.join(self.save_directory, 'data-file-schema.xml')
        if not os.path.exists(xmlfile):
            return None, {}
        try:
            atoms, xml_input = xml_parser(xmlfile)
            self.logger.debug('Read geometry and parameters from xml file successfully.')
        except:
            self.logger.debug('Read xml file failed.')
            return None, {}
        return atoms, xml_input
    def read_convergence(self):
        '''
        Read the status of the calculation.
        {
        '0': 'Done',
        '-1': ' manual cancelled',
        '1': 'Maximum CPU time exceeded', convergence NOT achieved after n iterations
             Then, change the parameter: 1) mixing_beta or 2) degauss
        '2': 'manual restart', 'Pending or not submit',
        '3': 'other errors',
        '4': 'unknow error',
        }
        '''
        # read the error file from queue
        # if self.queue:
        #     errfile = self.label + '.err'
        #     if not os.path.exists(errfile):
        #         self.logger.debug('%s not exists'%errfile)
        #         # return 2, 'Pending or not submit'
        #     else:
        #         errs = ['RESTART', 'pw.x', 'out-of-memory', 
        #                 'NODE FAILURE', 'TIME LIMIT', 'COMMUNICATION', 
        #                 'segmentation fault', 'PARSE_ERR', 'mpirun']
        #         cancelled = False
        #         with open(errfile, 'r') as f:
        #             lines = f.readlines()
        #             for line in lines:
        #                 if 'CANCELLED' in line:
        #                     cancelled = True
        #                 for err in errs:
        #                     if err in line:
        #                         self.logger.debug('Need restart')
        #                         return 3, line
        #             # cancelled by owner
        #             if len(lines) > 0:
        #                 if cancelled:
        #                     self.logger.debug('Cancelled')
        #                     return -1, lines[0]
        # no error from queue, or job not run in the queue
        output = self.label + '.pwo'
        if not os.path.exists(output):
            # print('%s not exists'%output)
            return 3, 'No pwo output file'
        with open(output, 'r') as f:
            lines = f.readlines()
            if len(lines) == 0: return 1, 'pwo file has nothing'
            stime = lines[1].split('starts on')[1]
            nlines = len(lines)
            n = min([200, nlines])
            lastlines = lines[-n:-1]
            for line in lastlines:
                if line.rfind('too many bands are not converged') > -1:
                    self.logger.debug('Need restart')
                    return 1, 'Reason: %s'%(line)
                if line.rfind('convergence NOT achieved after') > -1:
                    self.logger.debug('Need restart: %s'%line)
                    return 1, 'Reason: %s'%(line)
                if line.rfind('Maximum CPU time exceeded') > -1:
                    self.logger.debug('Need restart')
                    return 2, 'Reason: %s'%(line)
                if line.rfind('JOB DONE.') > -1:
                    self.logger.debug('JOB DONE')
                    return 0, line
            self.logger.debug('Not converged, %s'%lastlines[-1])
        return 4, line
    def read_convergence_post(self, package = 'pw'):
        '''
        Read the status of the calculation.
        {
        '0': 'Done',
        }
        '''
        
        output = self.label + '.%so'%package
        if not os.path.exists(output):
            # print('%s not exists'%output)
            return False, 'No pwo output file'
        with open(output, 'r') as f:
            lines = f.readlines()
            if len(lines) == 0: return False, 'pwo file has nothing'
            nlines = len(lines)
            n = min([100, nlines])
            for line in lines[-n:-1]:
                if line.rfind('JOB DONE.') > -1:
                    self.logger.debug('JOB DONE.')
                    return True, line
        return False, line
    def run(self, atoms = None, restart = False):
        '''
        run and restart
        '''
        if atoms is not None:
            self.atoms = atoms
        icount = 1
        converged = 4
        meg0 = 'None'
        self.atoms.calc = self
        if not restart:
            try:
                self.logger.debug('Run step: %s'%icount)
                self.atoms.get_potential_energy()
                converged, meg0 = self.read_convergence()
            except Exception as e:
                self.logger.debug('Not converge: %s'%e)
                converged, meg0 = self.read_convergence()
                icount += 1
                if converged == 0:
                    converged = 4
        # print(converged, xml_end, meg0)
        # restart 0, 1, 2
        exitfile = os.path.join(self.directory, 'EXIT')
        while converged > 0 or restart == 2:
            if os.path.exists(exitfile):
                self.logger.debug('Exit file exists!')
                os.remove(exitfile)
                sys.exit()
            restart = 1
            self.backup_file('%s.pwo'%self.prefix, directory = self.directory)
            self.backup_file('data-file-schema.xml', directory = self.save_directory)
            self.logger.debug('Run step: %s'%icount)
            xml_end = self.check_xml_file()
            if xml_end:
                self.parameters['input_data']['CONTROL']['restart_mode'] = 'restart'
            else:                
                self.parameters['input_data']['CONTROL']['restart_mode'] = 'from_scratch'
            if converged == 1:
                if 'electron_maxstep' in self.parameters['input_data']['ELECTRONS']:
                    electron_maxstep = int(self.parameters['input_data']['ELECTRONS']['electron_maxstep'])
                else:
                    electron_maxstep = 100
                if 'mixing_beta' in self.parameters['input_data']['ELECTRONS']:
                    mixing_beta = float(self.parameters['input_data']['ELECTRONS']['mixing_beta'])
                else:
                    mixing_beta = 0.7
                if electron_maxstep < 200:
                    self.parameters['input_data']['ELECTRONS']['electron_maxstep'] = electron_maxstep + 50
                    self.logger.debug('electron_maxstep change from %s to %s'%(electron_maxstep, electron_maxstep + 50))
                self.parameters['input_data']['ELECTRONS']['mixing_beta'] = mixing_beta/2.0
                self.logger.debug('mixing_beta change from %s to %s'%(mixing_beta, mixing_beta/2.0))
            try:
                icount += 1
                atoms.get_potential_energy()
                converged, meg = self.read_convergence()
            except Exception as e:
                self.logger.debug('Not converge: %s'%e)
                converged, meg = self.read_convergence()
                # meg = e
                restart = 2
                if converged == 0:
                    converged = 4
            self.logger.debug('%s, %s'%(converged, meg[0:-1]))
            if meg == meg0:
                self.logger.debug('Exit! Maybe deadblock! Same error message! %s'%meg)
                sys.exit()
            meg0 = meg
    def backup_file(self, src, directory = '.'):
        '''
        compare files
        backup
        '''
        import filecmp
        tnow = datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
        new_src = os.path.join(directory, '%s-%s'%(tnow, src))
        src = os.path.join(directory, src)
        if not os.path.exists(src): return 0
        ftype = src.split('.')[-1]
        flag = True
        for dst in os.listdir(directory):
            dst = os.path.join(directory, dst)
            if dst == src: continue
            if dst.endswith(ftype):
                if filecmp.cmp(dst, src):
                    flag = False
        # print('backup files: ', flag, src, new_src)
        if flag:
            shutil.copy(src, new_src)
        return 0
    def nscf(self, calculation = 'nscf', package = 'pw', queue = False, parallel = '', kpts = (10, 10, 10), **kwargs):
        import copy
        # save scf parameters
        # save scf parameters
        print("{0:=^60}".format(calculation))
        if not self.scf_directory:
            self.scf_directory = self.directory
        if not self.scf_parameters:
            self.scf_parameters = copy.deepcopy(self.parameters)
        if not self.scf_results:
            self.scf_results = copy.deepcopy(self.results)
            self.efermi = self.get_fermi_level()
        # read new atoms from results and set magmoms
        self.atoms = self.results['atoms']
        self.parameters = copy.deepcopy(self.scf_parameters)
        self.parameters['kpts'] = kpts
        self.parameters['input_data']['CONTROL']['calculation'] = calculation
        self.parameters['input_data']['CONTROL']['verbosity'] = 'high'
        self.parameters['input_data']['CONTROL']['outdir'] = '../'
        self.parameters['input_data']['SYSTEM']['occupations'] = 'tetrahedra'
        self.parameters['input_data']['SYSTEM'].pop('degauss', None)
        # self.parameters['restart_mode'] = 'restart'
        for key, value in kwargs.items():
            self.parameters[key] = value
        self.parameters = self.check_input(self.parameters)
        # create working directory
        self.directory = os.path.join(self.scf_directory, '%s/'%calculation)
        self.label = os.path.join(self.directory, self.prefix)
        self.nscf_asei = os.path.join(self.directory, '%s.nscf_asei'%self.prefix)
        # create working directory
        self.directory = os.path.join(self.scf_directory, '%s/'%calculation)
        self.label = os.path.join(self.directory, self.prefix)
        self.set_label(self.directory, self.prefix)
        self.set_queue(package = package, parallel = parallel, queue = queue)
    def nscf_calculate(self, ):
        from xespresso.utils import get_hash
        # read information of the charge-density file
        chargeFile = os.path.join(self.scf_directory, '%s.save/charge-density.dat'%self.prefix)
        self.state_info = get_hash(chargeFile)
        self.state_parameters = self.parameters
        output, meg = self.read_convergence_post('pw')
        if output:
            if os.path.isfile(self.nscf_asei):
                system_changes =  self.check_state_post(self.nscf_asei, package='PW')
                if not system_changes:
                    self.logger.debug('Use previous results!')
                    return 0
        write_espresso_asei(self.nscf_asei, self.state_info, self.state_parameters)
        self.calculate()
        # self.state_info = get_hash(chargeFile)
        # write_espresso_asei(self.nscf_asei, self.state_info, self.state_parameters)
        # self.read_results()

    
    def post(self, queue = False, package = 'dos', parallel = '', **kwargs):
        '''
        todo: 
        '''
        from xespresso.utils import get_hash
        print('{0:=^60}'.format(package))
        self.directory = os.path.join(self.scf_directory, '%s/'%package)
        self.set_label(self.directory, self.prefix)
        self.post_asei = os.path.join(self.directory, '%s.post_asei'%self.prefix)
        kwargs['prefix'] = self.prefix
        kwargs['outdir'] = '../'
        self.state_parameters = copy.deepcopy(kwargs)
        for wfc in ['wfc1', 'wfcdw1']:
            wfcFile = os.path.join(self.scf_directory, '%s.save/%s.dat'%(self.prefix, wfc))
            if os.path.isfile(wfcFile):
                self.state_info = get_hash(wfcFile)
        output, meg = self.read_convergence_post(package)
        if output:
            self.logger.debug('Previous calculation done.')
            if os.path.isfile(self.post_asei):
                system_changes =  self.check_state_post(self.post_asei, package)
                if not system_changes:
                    self.logger.debug('File and Parameters did not change. Use previous results!')
                    return 0
        self.post_write_input(package, **self.state_parameters)
        write_espresso_asei(self.post_asei, self.state_info, self.state_parameters)
        self.set_queue(package = package, parallel = parallel, queue = queue)
        self.post_calculate()
        self.post_read_results()
    def check_state_post(self, asei, package):
        old_state_info, old_state_parameters = read_espresso_asei(asei, package)
        if not self.state_info == old_state_info:
            self.logger.debug('File in save changed')
            return True
        elif not self.state_parameters == old_state_parameters:
            self.logger.debug('Parameters changed')
            return True
        else:
            return False
    def post_write_input(self, package, **kwargs):
        state_parameters = {
        'dos':  {'DOS': ['prefix', 'outdir', 'bz_sum', 'ngauss', 'degauss', 
                        'Emin', 'Emax', 'DeltaE', 'fildo', ]
                        },
        'pp':   {'INPUTPP': ['prefix', 'outdir', 'filplot', 'plot_num', 
                       'spin_component', 'spin_component', 'emin', 'emax', 
                       'delta_e', 'degauss_ldos', 'sample_bias', 'kpoint', 
                       'kband', 'lsign', 'spin_component', 'emin', 'emax', 
                       'spin_component', 'spin_component', 'spin_component', 
                       'spin_component'], 
                'PLOT': ['nfile', 'filepp', 'weight', 'iflag', 'output_format', 
                    'fileout', 'interpolation', 'e1', 'x0', 'nx', 'e1', 'e2', 
                    'x0', 'nx', 'ny', 'e1', 'e2', 'e3', 'x0', 'nx', 'ny', 'nz', 
                    'radius', 'nx', 'ny']
                    },
        'projwfc': {'PROJWFC': ['prefix', 'outdir', 'ngauss', 'degauss', 
                    'Emin', 'Emax', 'DeltaE', 'lsym', 'pawproj', 'filpdos', 'filproj', 
                    'lwrite_overlaps', 'lbinary_data', 'kresolveddos', 'tdosinboxes', 
                    'n_proj_boxes', 'irmin(3,n_proj_boxes)', 'irmax(3,n_proj_boxes)', 
                    'plotboxes', ]
                    },
        'hp': {'INPUTHP': ['prefix','outdir','iverbosity','max_seconds','nq1','nq2',
                           'nq3','skip_equivalence_q','determine_num_pert_only',
                           'find_atpert','docc_thr','skip_type','equiv_type',
                           'perturb_only_atom','start_q','last_q','sum_pertq',
                           'compute_hp','conv_thr_chi','thresh_init','ethr_nscf',
                           'niter_max','alpha_mix(i)','nmix','num_neigh','lmin','rmax',]},
        'bands': {'BANDS': ['prefix', 'outdir', 'filband', 'spin_component', 'lsigma', 
                            'lp', 'filp', 'lsym', 'no_overlap', 'plot_2d', 'firstk', 'lastk']
                            },
        'ph': {'INPUTPH': ['amass', 'outdir', 'prefix', 'niter_ph', 'tr2_ph', 
                           'alpha_mix(niter)', 'nmix_ph', 'verbosity', 'reduce_io', 
                           'max_seconds', 'fildyn', 'fildrho', 'fildvscf', 'epsil', 
                           'lrpa', 'lnoloc', 'trans', 'lraman', 'eth_rps', 'eth_ns', 
                           'dek', 'recover', 'low_directory_check', 'only_init', 'qplot', 
                           'q2d', 'q_in_band_form', 'electron_phonon', 'el_ph_nsigma', 
                           'el_ph_sigma', 'ahc_dir', 'ahc_nbnd', 'ahc_nbndskip', 
                           'skip_upperfan', 'lshift_q', 'zeu', 'zue', 'elop', 'fpol', 
                           'ldisp', 'nogg', 'asr', 'ldiag', 'lqdir', 'search_sym', 
                           'nq1', 'nq2', 'nq3', 'nk1', 'nk2', 'nk3', 'k1', 'k2', 'k3', 
                           'diagonalization', 'read_dns_bare', 'ldvscf_interpolate', 
                           'wpot_dir', 'do_long_range', 'do_charge_neutral', 'start_irr', 
                           'last_irr', 'nat_todo', 'modenum', 'start_q', 'last_q', 
                           'dvscf_star', 'drho_star', ],
                'LINE': ['xq', 'atom'],
                # 'qPointsSpecs': ['nqs', 'xq1', 'xq2', 'xq3', 'nq'],
                            },
        'dynmat': {'INPUT': ['fildyn', 'q', 'amass', 'asr', 'axis', 'lperm', 'lplasma', 'filout', 
                             'fileig', 'filmol', 'filxsf', 'loto_2d', 'el_ph_nsig', 'el_ph_sigma']
                  },
        'matdyn': {'INPUT': ['flfrc', 'asr', 'dos', 'nk1', 'nk2', 'nk3', 'deltaE', 'ndos', 'fldos', 
                             'flfrq', 'flvec', 'fleig', 'fldvn', 'at', 'l1', 'l2', 'l3', 'ntyp', 
                             'amass', 'readtau', 'fltau', 'la2F', 'q_in_band_form', 'q_in_cryst_coord', 
                             'eigen_similarity', 'fd', 'na_ifc', 'nosym', 'loto_2d', 
                             'loto_disable']
                  },
        'q2r': {'INPUT': ['fildyn', 'flfrc', 'zasr', 'loto_2d'],},
        }
        
        
        package_parameters = state_parameters[package]
        filename = os.path.join(self.directory, '%s.%si' %(self.prefix, package))
        with open(filename, 'w') as f:
            for section, parameters in package_parameters.items():
                if section != 'LINE':
                    f.write('&%s\n'%section)
                    for key, value in kwargs.items():
                        if key in parameters:
                            if isinstance(value, dict):
                                for subkey, subvalue in value.items():
                                    if isinstance(subvalue, str):
                                        f.write('  %s(%s) = "%s", \n' %(key, subkey, subvalue))
                                    else:
                                        f.write('  %s(%s) = %s, \n' %(key, subkey, subvalue))
                            else:
                                if isinstance(value, str):
                                    f.write('  {0:10s} =  "{1}" \n'.format(key, value))
                                else:
                                    f.write('  {0:10s} =  {1} \n'.format(key, value))
                    f.write('/ \n')
                else:
                    for key, value in kwargs.items():
                        if key in parameters:
                            f.write('  %s \n' %(value))

    def post_calculate(self):
        command = self.command
        print('Running %s'%self.package)
        try:
            proc = subprocess.Popen(command, shell=True, cwd=self.directory)
        except OSError as err:
            msg = 'Failed to execute "{}"'.format(command)
            raise EnvironmentError(msg) from err

        errorcode = proc.wait()

        if errorcode:
            path = os.path.abspath(self.directory)
            msg = ('Command "{}" failed in '
                   '{} with error code {}'.format(command,
                                                  path, errorcode))
            raise CalculationFailed(msg)
        print('Done: %s' %self.package)
    def post_read_results(self):
        '''
        '''
        pass
    def plot_phdos(self, fldos = None, ax = None, output = None):
        '''
        '''
        import matplotlib.pyplot as plt
        if fldos is None:
            fldos = self.prefix
        phdos = np.loadtxt(self.directory+'/%s.phdos' % fldos)
        self.phdos = phdos
        if ax is None:
            fig, ax = plt.subplots(figsize = (6, 3))
            # ax = plt.gca()
        ax.plot(self.phdos[:, 0], self.phdos[:, 1], linewidth=0.7)
        ax.set_xlabel('Frequency (cm^-1)')
        ax.set_ylabel('DOS (a.u.)')
        if output is not None:
            plt.savefig('%s'%output)
        return ax
    def get_work_function(self, ax = None, inpfile = 'potential.cube', output = None, shift = False):
        import matplotlib.pyplot as plt
        from ase.io.cube import read_cube_data
        from ase.units import create_units
        if ax is None:
            import matplotlib.pyplot as plt
            ax = plt.gca()
        units = create_units('2006')
        #
        filename = os.path.join(self.directory, inpfile)
        data, atoms = read_cube_data(filename)
        data = data * units['Ry']
        ef = self.get_fermi_level()
        # x, y, z, lp = calc.get_local_potential()
        nx, ny, nz = data.shape
        axy = np.array([np.average(data[:, :, z]) for z in range(nz)])
        # setup the x-axis in realspace
        uc = atoms.get_cell()
        xaxis = np.linspace(0, uc[2][2], nz)
        if shift:
            ax.plot(xaxis, axy - ef)
            ef = 0
        else:
            ax.plot(xaxis, axy)
            ax.plot([min(xaxis), max(xaxis)], [ef, ef], 'k:')
        ax.set_xlabel('Position along z-axis ($\AA$)')
        ax.set_ylabel('Potential (eV)')
        if output:
            plt.savefig('%s'%output)
        # plt.show()
        atoms = self.results['atoms']
        pos = max(atoms.positions[:, 2] + 3)
        ind = (xaxis > pos) & (xaxis < pos + 3)
        wf = np.average(axy[ind]) - ef
        print('min: %s, max: %s'%(pos, pos + 3))
        print('The workfunction is {0:1.2f} eV'.format(wf))
    def get_bader_charge(self, inpfile = None):
        '''
        '''
        from ase.io.bader import attach_charges
        if not inpfile:
            inpfile = '%s.cube' %self.prefix
        command = 'bader %s'%inpfile
        print(command)
        try:
            proc = subprocess.Popen(command, shell=True, cwd=self.directory)
        except OSError as err:
            msg = 'Failed to execute "{}"'.format(command)
            raise EnvironmentError(msg) from err
        acf = os.path.join(self.directory, 'ACF.dat')
        attach_charges(self.results['atoms'], acf)
    def read_time(self, ):
        '''
        '''
        pwo = os.path.join(self.directory, self.prefix + '.pwo')
        with open(pwo, 'r') as f:
            ts = 0
            lines = f.readlines()
            for line in lines[::-1]:
                if 'PWSCF' in line and 'WALL' in line:
                    t = line.split('CPU')[-1].split('WALL')[0]
                    if 's' in t:
                        ts = float(t.split('m')[-1].split('s')[0])
                    if 'm' in t:
                        tm = float(t.split('h')[-1].split('m')[0])
                        ts += tm*60
                    if 'h' in t:
                        th= float(t.split('h')[0])
                        ts += th*3600
                    break
                self.results['time'] = ts
        return ts
    def get_time(self, ):
        t = 0
        filename = os.path.join(self.directory, '%s.pwo'%self.prefix)
        with open(filename) as f:
            lines = f.readlines()
            for line in lines[-200:]:
                if 'init_run     :' in line or 'electrons    :' in line:
                    t += float(line.split('CPU')[1].split('s WALL')[0])
        self.results['time'] = t
        return t
    def clean(self):
        '''
        remove wfc, hub files
        '''
        keys = ['.wfc', '.hub']
        files = os.listdir(self.directory)
        for file in files:
            for key in keys:
                if key in file:
                    os.remove(os.path.join(self.directory, file))
        files = os.listdir(self.save_directory)
        # keys = ['wfc']
        # for file in files:
        #     for key in keys:
        #         if key in file:
        #             os.remove(os.path.join(self.save_directory, file))
    def get_fermi_level(self):
        if self.calc is None:
            raise PropertyNotPresent(error_template % 'Fermi level')
        return self.calc.get_fermi_level()

    def get_ibz_k_points(self):
        if self.calc is None:
            raise PropertyNotPresent(error_template % 'IBZ k-points')
        ibzkpts = self.calc.get_ibz_k_points()
        if ibzkpts is None:
            warnings.warn(warn_template % 'IBZ k-points')
        return ibzkpts

    def get_k_point_weights(self):
        if self.calc is None:
            raise PropertyNotPresent(error_template % 'K-point weights')
        k_point_weights = self.calc.get_k_point_weights()
        if k_point_weights is None:
            warnings.warn(warn_template % 'K-point weights')
        return k_point_weights

    def get_eigenvalues(self, **kwargs):
        if self.calc is None:
            raise PropertyNotPresent(error_template % 'Eigenvalues')
        eigenvalues = self.calc.get_eigenvalues(**kwargs)
        if eigenvalues is None:
            warnings.warn(warn_template % 'Eigenvalues')
        return eigenvalues

    def get_number_of_spins(self):
        if self.calc is None:
            raise PropertyNotPresent(error_template % 'Number of spins')
        nspins = self.calc.get_number_of_spins()
        if nspins is None:
            warnings.warn(warn_template % 'Number of spins')
        return nspins
    





restart_ignore = {'PW': ['max_seconds', 'nstep', 'restart_mode',
'electron_maxstep', 
'scf_must_converge', 
'mixing_mode', 
'mixing_beta', 
'mixing_ndim', 
'mixing_fixed_ns', 
'diagonalization', 
'diago_cg_maxiter', 
'diago_david_ndim', 
'diago_full_acc', 
'startingpot', 
'startingwfc', 
'tqr', 
'real_space', 
'title',
'disk_io', 
'wf_collect', 
'outdir',
'pseudo_dir',
#
'diago_thr_init',
'tprnfor', 
]}

default = {
'PW': {'disk_io': {'scf': 'low', 'nscf': 'medium', 'bands': 'medium', 'relax': 'medium', 'md': 'medium', 'vc-relax': 'medium', 'vc-md': 'medium'},
       'nstep': {'scf': 1, 'nscf': 1, 'bands': 1, 'relax': 50, 'md': 50, 'vc-relax': 50, 'vc-md': 50},
       'pseudo_dir': '$HOME/espresso/pseudo/',
       },
}

Ha2Ry = {
'ecutwfc',
'ecutrho',
'ecutfock', 
# 'degauss',  # not sure
'etot_conv_thr',
'forc_conv_thr',
}

qe_namespace = {'BANDS': {'BANDS': {'filband': ['CHARACTER', []],
                      'filp': ['CHARACTER', []],
                      'firstk, lastk': ['INTEGER', []],
                      'lp': ['LOGICAL', []],
                      'lsigma(i), i=1,3': ['LOGICAL', []],
                      'lsym': ['LOGICAL', []],
                      'no_overlap': ['LOGICAL', []],
                      'outdir': ['CHARACTER', []],
                      'plot_2d': ['LOGICAL', []],
                      'prefix': ['CHARACTER', []],
                      'spin_component': ['INTEGER', []]}},
 'DOS': {'DOS': {'DeltaE': ['REAL', []],
                  'Emin, Emax': ['REAL', []],
                  'bz_sum': ['CHARACTER',
                             ['smearing',
                              'tetrahedra',
                              'tetrahedra_lin',
                              'tetrahedra_opt']],
                  'degauss': ['REAL', []],
                  'fildos': ['CHARACTER', []],
                  'ngauss': ['INTEGER', []],
                  'outdir': ['CHARACTER', []],
                  'prefix': ['CHARACTER', []]}},
 'NEB': {'PATH': {'CI_scheme': ['CHARACTER', ['no-CI', 'auto', 'manual']],
                   'ds': ['REAL', []],
                   'fcp_mu': ['REAL', []],
                   'fcp_tot_charge_first': ['REAL', []],
                   'fcp_tot_charge_last': ['REAL', []],
                   'first_last_opt': ['LOGICAL', []],
                   'k_max, k_min': ['REAL', []],
                   'lfcpopt': ['LOGICAL', []],
                   'minimum_image': ['LOGICAL', []],
                   'nstep_path': ['INTEGER', []],
                   'num_of_images': ['INTEGER', []],
                   'opt_scheme': ['CHARACTER',
                                  ['sd',
                                   'broyden',
                                   'broyden2',
                                   'quick-min',
                                   'langevin']],
                   'path_thr': ['REAL', []],
                   'restart_mode': ['CHARACTER', ['from_scratch', 'restart']],
                   'string_method': ['CHARACTER', ['neb', 'smd']],
                   'temp_req': ['REAL', []],
                   'use_freezing': ['LOGICAL', []],
                   'use_masses': ['LOGICAL', []]},
         'ATOMIC_POSITIONS': {},
         'CLIMBING_IMAGES': {'index1, index2, ... indexN': ['INTEGER', []]}},
 'PP': {'INPUTPP': {'degauss_ldos': ['REAL', []],
                     'delta_e': ['REAL', []],
                     'emax': ['REAL', []],
                     'emin': ['REAL', []],
                     'filplot': ['CHARACTER', []],
                     'kband(i), i=1,2': ['INTEGER', []],
                     'kpoint(i), i=1,2': ['INTEGER', []],
                     'lsign': ['LOGICAL', []],
                     'outdir': ['CHARACTER', []],
                     'plot_num': ['INTEGER', []],
                     'prefix': ['CHARACTER', []],
                     'sample_bias': ['REAL', []],
                     'spin_component': ['INTEGER', []],
                     'spin_component(i), i=1,2': ['INTEGER', []]},
        'PLOT': {'e1(i), \n\t\te2(i), \n\t\t \n\t\t  i=1,3': ['REAL', []],
                  'e1(i), \n\t\te2(i), \n\t\te3(i), \n\t\t \n\t\t  i=1,3': ['REAL',
                                                                            []],
                  'e1(i), i=1,3': ['REAL', []],
                  'fileout': ['CHARACTER', []],
                  'filepp(i), i=1,nfile': ['CHARACTER', []],
                  'iflag': ['INTEGER', []],
                  'interpolation': ['CHARACTER', ['fourier', 'bspline']],
                  'nfile': ['INTEGER', []],
                  'nx': ['INTEGER', []],
                  'nx, ny': ['INTEGER', []],
                  'nx, ny, nz': ['INTEGER', []],
                  'output_format': ['INTEGER', []],
                  'radius': ['REAL', []],
                  'weight(i), i=1,nfile': ['REAL', []],
                  'x0(i), i=1,3': ['REAL', []]}},
 'PROJWFC': {'PROJWFC': {'DeltaE': ['REAL', []],
                          'Emin, Emax': ['REAL', []],
                          'degauss': ['REAL', []],
                          'filpdos': ['CHARACTER', []],
                          'filproj': ['CHARACTER', []],
                          'irmax(3,n_proj_boxes)': ['INTEGER', []],
                          'irmin(3,n_proj_boxes)': ['INTEGER', []],
                          'kresolveddos': ['LOGICAL', []],
                          'lbinary_data': ['LOGICAL', []],
                          'lsym': ['LOGICAL', []],
                          'lwrite_overlaps': ['LOGICAL', []],
                          'n_proj_boxes': ['INTEGER', []],
                          'ngauss': ['INTEGER', []],
                          'outdir': ['CHARACTER', []],
                          'pawproj': ['LOGICAL', []],
                          'plotboxes': ['LOGICAL', []],
                          'prefix': ['CHARACTER', []],
                          'tdosinboxes': ['LOGICAL', []]}},
 'PW': {'CELL': {'cell_dofree': ['CHARACTER',
                                  ['all',
                                   'ibrav',
                                   'x',
                                   'y',
                                   'z',
                                   'xy',
                                   'xz',
                                   'yz',
                                   'xyz',
                                   'shape',
                                   'volume',
                                   '2Dxy',
                                   '2Dshape',
                                   'epitaxial_ab',
                                   'epitaxial_ac',
                                   'epitaxial_bc']],
                  'cell_dynamics': ['CHARACTER',
                                    ['none',
                                     'sd',
                                     'damp-pr',
                                     'damp-w',
                                     'bfgs',
                                     'none',
                                     'pr',
                                     'w']],
                  'cell_factor': ['REAL', []],
                  'press': ['REAL', []],
                  'press_conv_thr': ['REAL', []],
                  'wmass': ['REAL', []]},
        'CONTROL': {'calculation': ['CHARACTER',
                                     ['scf',
                                      'nscf',
                                      'bands',
                                      'relax',
                                      'md',
                                      'vc-relax',
                                      'vc-md']],
                     'dipfield': ['LOGICAL', []],
                     'disk_io': ['CHARACTER',
                                 ['high', 'medium', 'low', 'nowf', 'none']],
                     'dt': ['REAL', []],
                     'etot_conv_thr': ['REAL', []],
                     'forc_conv_thr': ['REAL', []],
                     'gate': ['LOGICAL', []],
                     'gdir': ['INTEGER', []],
                     'iprint': ['INTEGER', []],
                     'lberry': ['LOGICAL', []],
                     'lelfield': ['LOGICAL', []],
                     'lfcpopt': ['LOGICAL', []],
                     'lkpoint_dir': ['LOGICAL', []],
                     'lorbm': ['LOGICAL', []],
                     'max_seconds': ['REAL', []],
                     'nberrycyc': ['INTEGER', []],
                     'nppstr': ['INTEGER', []],
                     'nstep': ['INTEGER', []],
                     'outdir': ['CHARACTER', []],
                     'prefix': ['CHARACTER', []],
                     'pseudo_dir': ['CHARACTER', []],
                     'restart_mode': ['CHARACTER', ['from_scratch', 'restart']],
                     'tefield': ['LOGICAL', []],
                     'title': ['CHARACTER', []],
                     'tprnfor': ['LOGICAL', []],
                     'tstress': ['LOGICAL', []],
                     'verbosity': ['CHARACTER', ['high', 'low']],
                     'wf_collect': ['LOGICAL', []],
                     'wfcdir': ['CHARACTER', []]},
        'ELECTRONS': {'adaptive_thr': ['LOGICAL', []],
                       'conv_thr': ['REAL', []],
                       'conv_thr_init': ['REAL', []],
                       'conv_thr_multi': ['REAL', []],
                       'diago_cg_maxiter': ['INTEGER', []],
                       'diago_david_ndim': ['INTEGER', []],
                       'diago_full_acc': ['LOGICAL', []],
                       'diago_thr_init': ['REAL', []],
                       'diagonalization': ['CHARACTER',
                                           ['david',
                                            'cg',
                                            'ppcg',
                                            'paro',
                                            'ParO']],
                       'efield': ['REAL', []],
                       'efield_cart(i), i=1,3': ['REAL', []],
                       'efield_phase': ['CHARACTER', ['read', 'write', 'none']],
                       'electron_maxstep': ['INTEGER', []],
                       'mixing_beta': ['REAL', []],
                       'mixing_fixed_ns': ['INTEGER', []],
                       'mixing_mode': ['CHARACTER',
                                       ['plain', 'TF', 'local-TF']],
                       'mixing_ndim': ['INTEGER', []],
                       'real_space': ['LOGICAL', []],
                       'scf_must_converge': ['LOGICAL', []],
                       'startingpot': ['CHARACTER', ['atomic', 'file']],
                       'startingwfc': ['CHARACTER',
                                       ['atomic',
                                        'atomic+random',
                                        'random',
                                        'file']],
                       'tqr': ['LOGICAL', []]},
        'IONS': {'bfgs_ndim': ['INTEGER', []],
                  'delta_t': ['REAL', []],
                  'ion_dynamics': ['CHARACTER',
                                   ['bfgs',
                                    'damp',
                                    'verlet',
                                    'langevin',
                                    'langevin-smc',
                                    'bfgs',
                                    'damp',
                                    'beeman']],
                  'ion_positions': ['CHARACTER', ['default', 'from_input']],
                  'ion_temperature': ['CHARACTER',
                                      ['rescaling',
                                       'rescale-v',
                                       'rescale-T',
                                       'reduce-T',
                                       'berendsen',
                                       'andersen',
                                       'svr',
                                       'initial',
                                       'not_controlled']],
                  'ion_velocities': ['CHARACTER', ['default', 'from_input']],
                  'nraise': ['INTEGER', []],
                  'pot_extrapolation': ['CHARACTER',
                                        ['none',
                                         'atomic',
                                         'first_order',
                                         'second_order']],
                  'refold_pos': ['LOGICAL', []],
                  'remove_rigid_rot': ['LOGICAL', []],
                  'tempw': ['REAL', []],
                  'tolp': ['REAL', []],
                  'trust_radius_ini': ['REAL', []],
                  'trust_radius_max': ['REAL', []],
                  'trust_radius_min': ['REAL', []],
                  'upscale': ['REAL', []],
                  'w_1': ['REAL', []],
                  'w_2': ['REAL', []],
                  'wfc_extrapolation': ['CHARACTER',
                                        ['none',
                                         'first_order',
                                         'second_order']]},
        'SYSTEM': {'A, B, C, cosAB, cosAC, cosBC': ['REAL', []],
                    'Hubbard_J(i,ityp), (i,ityp) = (1,1) ... (3,ntyp)': ['REAL',
                                                                         []],
                    'Hubbard_J0(i), i=1,ntype': ['REAL', []],
                    'Hubbard_U(i), i=1,ntyp': ['REAL', []],
                    'Hubbard_V(na,nb,k), (na,nb,k) = (1,1,1) ... (natx,27*natx,4)': ['REAL',
                                                                                     []],
                    'Hubbard_alpha(i), i=1,ntyp': ['REAL', []],
                    'Hubbard_beta(i), i=1,ntyp': ['REAL', []],
                    'Hubbard_parameters': ['CHARACTER', ['input', 'file']],
                    'U_projection_type': ['CHARACTER',
                                          ['atomic',
                                           'ortho-atomic',
                                           'norm-atomic',
                                           'file',
                                           'pseudo']],
                    'ace': ['LOGICAL', []],
                    'angle1(i), i=1,ntyp': ['REAL', []],
                    'angle2(i), i=1,ntyp': ['REAL', []],
                    'assume_isolated': ['CHARACTER',
                                        ['none',
                                         'makov-payne',
                                         'm-p',
                                         'mp',
                                         'martyna-tuckerman',
                                         'm-t',
                                         'mt',
                                         'esm',
                                         '2D']],
                    'block': ['LOGICAL', []],
                    'block_1': ['REAL', []],
                    'block_2': ['REAL', []],
                    'block_height': ['REAL', []],
                    'celldm(i), i=1,6': ['REAL', []],
                    'constrained_magnetization': ['CHARACTER', []],
                    'degauss': ['REAL', []],
                    'dftd3_threebody': ['LOGICAL', []],
                    'dftd3_version': ['integer',
                                      ['dftd3_version=2',
                                       'dftd3_version=3',
                                       'dftd3_version=4',
                                       'dftd3_version=5',
                                       'dftd3_version=6']],
                    'eamp': ['REAL', []],
                    'ecfixed': ['REAL', []],
                    'ecutfock': ['REAL', []],
                    'ecutrho': ['REAL', []],
                    'ecutvcut': ['REAL', []],
                    'ecutwfc': ['REAL', []],
                    'edir': ['INTEGER', []],
                    'emaxpos': ['REAL', []],
                    'ensemble_energies': ['LOGICAL', []],
                    'eopreg': ['REAL', []],
                    'esm_bc': ['CHARACTER', []],
                    'esm_efield': ['REAL', []],
                    'esm_nfit': ['INTEGER', []],
                    'esm_w': ['REAL', []],
                    'exx_fraction': ['REAL', []],
                    'exxdiv_treatment': ['CHARACTER',
                                         ['gygi-baldereschi',
                                          'vcut_spherical',
                                          'vcut_ws',
                                          'none']],
                    'fcp_mu': ['REAL', []],
                    'fixed_magnetization(i), i=1,3': ['REAL', []],
                    'force_symmorphic': ['LOGICAL', []],
                    'ibrav': ['INTEGER', []],
                    'input_dft': ['CHARACTER', []],
                    'lambda': ['REAL', []],
                    'lda_plus_u': ['LOGICAL', []],
                    'lda_plus_u_kind': ['INTEGER', []],
                    'lforcet': ['LOGICAL', []],
                    'localization_thr': ['REAL', []],
                    'london': ['LOGICAL', []],
                    'london_c6(i), i=1,ntyp': ['REAL', []],
                    'london_rcut': ['REAL', []],
                    'london_rvdw(i), i=1,ntyp': ['REAL', []],
                    'london_s6': ['REAL', []],
                    'lspinorb': ['LOGICAL', []],
                    'nat': ['INTEGER', []],
                    'nbnd': ['INTEGER', []],
                    'no_t_rev': ['LOGICAL', []],
                    'noinv': ['LOGICAL', []],
                    'noncolin': ['LOGICAL', []],
                    'nosym': ['LOGICAL', []],
                    'nosym_evc': ['LOGICAL', []],
                    'nqx1, nqx2, nqx3': ['INTEGER', []],
                    'nr1, nr2, nr3': ['INTEGER', []],
                    'nr1s, nr2s, nr3s': ['INTEGER', []],
                    'nspin': ['INTEGER', []],
                    'ntyp': ['INTEGER', []],
                    'occupations': ['CHARACTER',
                                    ['smearing',
                                     'tetrahedra',
                                     'tetrahedra_lin',
                                     'tetrahedra_opt',
                                     'fixed',
                                     'from_input']],
                    'one_atom_occupations': ['LOGICAL', []],
                    'origin_choice': ['INTEGER', []],
                    'q2sigma': ['REAL', []],
                    'qcutz': ['REAL', []],
                    'relaxz': ['LOGICAL', []],
                    'report': ['INTEGER', []],
                    'rhombohedral': ['LOGICAL', []],
                    'screening_parameter': ['REAL', []],
                    'smearing': ['CHARACTER',
                                 ['gaussian',
                                  'gauss',
                                  'methfessel-paxton',
                                  'm-p',
                                  'mp',
                                  'marzari-vanderbilt',
                                  'cold',
                                  'm-v',
                                  'mv',
                                  'fermi-dirac',
                                  'f-d',
                                  'fd']],
                    'space_group': ['INTEGER', []],
                    'starting_charge(i), i=1,ntyp': ['REAL', []],
                    'starting_magnetization(i), i=1,ntyp': ['REAL', []],
                    'starting_ns_eigenvalue(m,ispin,ityp), (m,ispin,ityp) = (1,1,1) ... (2*lmax+1,nspin or npol,ntyp)': ['REAL',
                                                                                                                         []],
                    'starting_spin_angle': ['LOGICAL', []],
                    'tot_charge': ['REAL', []],
                    'tot_magnetization': ['REAL', []],
                    'ts_vdw_econv_thr': ['REAL', []],
                    'ts_vdw_isolated': ['LOGICAL', []],
                    'uniqueb': ['LOGICAL', []],
                    'use_all_frac': ['LOGICAL', []],
                    'vdw_corr': ['CHARACTER', []],
                    'x_gamma_extrapolation': ['LOGICAL', []],
                    'xdm': ['LOGICAL', []],
                    'xdm_a1': ['REAL', []],
                    'xdm_a2': ['REAL', []],
                    'zgate': ['REAL', []]},
        'ATOMIC_FORCES': {'X': ['CHARACTER', []], 'fx, fy, fz': ['REAL', []]},
        'ATOMIC_POSITIONS': {"Card's options:": ["Card's options:",
                                                 ['alat',
                                                  'bohr',
                                                  'angstrom',
                                                  'crystal',
                                                  'crystal_sg']],
                             'X': ['CHARACTER', []],
                             'if_pos(1), if_pos(2), if_pos(3)': ['INTEGER', []],
                             'x, y, z': ['REAL', []]},
        'ATOMIC_SPECIES': {'Mass_X': ['REAL', []],
                           'PseudoPot_X': ['CHARACTER', []],
                           'X': ['CHARACTER', []]},
        'ATOMIC_VELOCITIES': {'V': ['CHARACTER', []],
                              'vx, vy, vz': ['REAL', []]},
        'CELL_PARAMETERS': {"Card's options:": ["Card's options:", []],
                            'v1, v2, v3': ['REAL', []]},
        'CONSTRAINTS': {'constr(1), constr(2), constr(3), constr(4)': ['', []],
                        'constr_target': ['REAL', []],
                        'constr_tol': ['REAL', []],
                        'constr_type': ['CHARACTER',
                                        ['type_coord',
                                         'atom_coord',
                                         'distance',
                                         'planar_angle',
                                         'torsional_angle',
                                         'bennett_proj']],
                        'nconstr': ['INTEGER', []]},
        'K_POINTS': {"Card's options:": ["Card's options:",
                                         ['tpiba',
                                          'automatic',
                                          'crystal',
                                          'gamma',
                                          'tpiba_b',
                                          'crystal_b',
                                          'tpiba_c',
                                          'crystal_c']],
                     'nk1, nk2, nk3': ['INTEGER', []],
                     'nks': ['INTEGER', []],
                     'sk1, sk2, sk3': ['INTEGER', []],
                     'xk_x, xk_y, xk_z, wk': ['REAL', []]},
        'OCCUPATIONS': {'f_inp1': ['REAL', []], 'f_inp2': ['REAL', []]}}}

default_parameters = {'BANDS': {'BANDS': {'filband': 'bands.out',
                      'filp': 'p_avg.dat',
                      'firstk, lastk': 1e+30,
                      'lp': False,
                      'lsigma(i), i=1,3': True,
                      'lsym': True,
                      'no_overlap': True,
                      'outdir': 'valueoftheESPRESSO_TMPDIRenvironmentvariableifset;\n'
                                'currentdirectory(./)otherwise',
                      'plot_2d': False,
                      'prefix': 'pwscf',
                      'spin_component': 1e+30}},
 'DOS': {'DOS': {'DeltaE': 1e+30,
                  'Emin, Emax': 1e+30,
                  'bz_sum': 'smearingifdegaussisgivenininput;\n'
                            'optionsreadfromthexmldatafileotherwise.',
                  'degauss': 1e+30,
                  'fildos': 'prefix.dos',
                  'ngauss': 1e+30,
                  'outdir': 'valueoftheESPRESSO_TMPDIRenvironmentvariableifset;\n'
                            'currentdirectory(./)otherwise',
                  'prefix': 'pwscf'}},
 'NEB': {'PATH': {'CI_scheme': 'no-CI',
                   'ds': 1.0,
                   'fcp_mu': 1e+30,
                   'fcp_tot_charge_first': 1e+30,
                   'fcp_tot_charge_last': 1e+30,
                   'first_last_opt': False,
                   'k_max, k_min': 1e+30,
                   'lfcpopt': True,
                   'minimum_image': False,
                   'nstep_path': 1,
                   'num_of_images': 0,
                   'opt_scheme': 'quick-min',
                   'path_thr': 1e+30,
                   'restart_mode': 'from_scratch',
                   'string_method': 'neb',
                   'temp_req': 1e+30,
                   'use_freezing': False,
                   'use_masses': False},
         'ATOMIC_POSITIONS': {},
         'CLIMBING_IMAGES': {'index1, index2, ... indexN': 1e+30}},
 'PP': {'INPUTPP': {'degauss_ldos': 1e+30,
                     'delta_e': 1e+30,
                     'emax': 1e+30,
                     'emin': 1e+30,
                     'filplot': '',
                     'kband(i), i=1,2': 1e+30,
                     'kpoint(i), i=1,2': 1e+30,
                     'lsign': True,
                     'outdir': 'valueoftheESPRESSO_TMPDIRenvironmentvariableifset;\n'
                               'currentdirectory(./)otherwise',
                     'plot_num': 1e+30,
                     'prefix': '',
                     'sample_bias': 1e+30,
                     'spin_component': 0,
                     'spin_component(i), i=1,2': 1e+30},
        'PLOT': {'e1(i), \n\t\te2(i), \n\t\t \n\t\t  i=1,3': 1e+30,
                  'e1(i), \n\t\te2(i), \n\t\te3(i), \n\t\t \n\t\t  i=1,3': 1e+30,
                  'e1(i), i=1,3': 1e+30,
                  'fileout': 'standardoutput',
                  'filepp(i), i=1,nfile': 'filepp(1)=filplot',
                  'iflag': 1e+30,
                  'interpolation': 'fourier',
                  'nfile': 1e+30,
                  'nx': 1e+30,
                  'nx, ny': 1e+30,
                  'nx, ny, nz': 1e+30,
                  'output_format': 1e+30,
                  'radius': 1e+30,
                  'weight(i), i=1,nfile': 1e+30,
                  'x0(i), i=1,3': 1e+30}},
 'PROJWFC': {'PROJWFC': {'DeltaE': 1e+30,
                          'Emin, Emax': 1e+30,
                          'degauss': 0.0,
                          'filpdos': '(valueofprefixvariable)',
                          'filproj': '(standardoutput)',
                          'irmax(3,n_proj_boxes)': 1e+30,
                          'irmin(3,n_proj_boxes)': 1e+30,
                          'kresolveddos': False,
                          'lbinary_data': False,
                          'lsym': True,
                          'lwrite_overlaps': False,
                          'n_proj_boxes': 1,
                          'ngauss': 0,
                          'outdir': 'valueoftheESPRESSO_TMPDIRenvironmentvariableifset;\n'
                                    'currentdirectory(./)otherwise',
                          'pawproj': False,
                          'plotboxes': False,
                          'prefix': 'pwscf',
                          'tdosinboxes': False}},
 'PW': {'CELL': {'cell_dofree': 'all',
                  'cell_dynamics': '',
                  'cell_factor': 1e+30,
                  'press': 0.0,
                  'press_conv_thr': 1e+30,
                  'wmass': 1e+30},
        'CONTROL': {'calculation': 'scf',
                     'dipfield': False,
                     'disk_io': 'seebelow',
                     'dt': 20.0,
                     'etot_conv_thr': 0.0001,
                     'forc_conv_thr': 0.001,
                     'gate': True,
                     'gdir': 1e+30,
                     'iprint': 1e+30,
                     'lberry': False,
                     'lelfield': False,
                     'lfcpopt': True,
                     'lkpoint_dir': True,
                     'lorbm': False,
                     'max_seconds': 1e+30,
                     'nberrycyc': 1,
                     'nppstr': 1e+30,
                     'nstep': 1e+30,
                     'outdir': 'valueoftheESPRESSO_TMPDIRenvironmentvariableifset;\n'
                               'currentdirectory(./)otherwise',
                     'prefix': 'pwscf',
                     'pseudo_dir': 'valueofthe$ESPRESSO_PSEUDOenvironmentvariableifset;\n'
                                   '$HOME/espresso/pseudo/otherwise',
                     'restart_mode': 'from_scratch',
                     'tefield': False,
                     'title': '',
                     'tprnfor': True,
                     'tstress': False,
                     'verbosity': 'low',
                     'wf_collect': True,
                     'wfcdir': 'sameasoutdir'},
        'ELECTRONS': {'adaptive_thr': False,
                       'conv_thr': 1e-06,
                       'conv_thr_init': 0.001,
                       'conv_thr_multi': 0.1,
                       'diago_cg_maxiter': 1e+30,
                       'diago_david_ndim': 2,
                       'diago_full_acc': False,
                       'diago_thr_init': 1e+30,
                       'diagonalization': 'david',
                       'efield': 0.0,
                       'efield_cart(i), i=1,3': 1e+30,
                       'efield_phase': 'none',
                       'electron_maxstep': 100,
                       'mixing_beta': 0.7,
                       'mixing_fixed_ns': 0,
                       'mixing_mode': 'plain',
                       'mixing_ndim': 8,
                       'real_space': False,
                       'scf_must_converge': True,
                       'startingpot': '',
                       'startingwfc': 'atomic+random',
                       'tqr': False},
        'IONS': {'bfgs_ndim': 1,
                  'delta_t': 1.0,
                  'ion_dynamics': '',
                  'ion_positions': 'default',
                  'ion_temperature': 'not_controlled',
                  'ion_velocities': 'default',
                  'nraise': 1,
                  'pot_extrapolation': 'atomic',
                  'refold_pos': False,
                  'remove_rigid_rot': False,
                  'tempw': 300.0,
                  'tolp': 100.0,
                  'trust_radius_ini': 0.5,
                  'trust_radius_max': 0.8,
                  'trust_radius_min': 0.001,
                  'upscale': 100.0,
                  'w_1': 0.01,
                  'w_2': 0.5,
                  'wfc_extrapolation': 'none'},
        'SYSTEM': {'A, B, C, cosAB, cosAC, cosBC': 1e+30,
                    'Hubbard_J(i,ityp), (i,ityp) = (1,1) ... (3,ntyp)': 1e+30,
                    'Hubbard_J0(i), i=1,ntype': 1e+30,
                    'Hubbard_U(i), i=1,ntyp': 1e+30,
                    'Hubbard_V(na,nb,k), (na,nb,k) = (1,1,1) ... (natx,27*natx,4)': 1e+30,
                    'Hubbard_alpha(i), i=1,ntyp': 1e+30,
                    'Hubbard_beta(i), i=1,ntyp': 1e+30,
                    'Hubbard_parameters': 'input',
                    'U_projection_type': 'atomic',
                    'ace': True,
                    'angle1(i), i=1,ntyp': 1e+30,
                    'angle2(i), i=1,ntyp': 1e+30,
                    'assume_isolated': 'none',
                    'block': False,
                    'block_1': 0.45,
                    'block_2': 0.55,
                    'block_height': 0.1,
                    'celldm(i), i=1,6': 1e+30,
                    'constrained_magnetization': '',
                    'degauss': 1e+30,
                    'dftd3_threebody': True,
                    'dftd3_version': 3,
                    'eamp': 1e+30,
                    'ecfixed': 0.0,
                    'ecutfock': 1e+30,
                    'ecutrho': 1e+30,
                    'ecutvcut': 1e+30,
                    'ecutwfc': 1e+30,
                    'edir': 1e+30,
                    'emaxpos': 0.5,
                    'ensemble_energies': False,
                    'eopreg': 0.1,
                    'esm_bc': '',
                    'esm_efield': 1e+30,
                    'esm_nfit': 1e+30,
                    'esm_w': 1e+30,
                    'exx_fraction': 1e+30,
                    'exxdiv_treatment': 'gygi-baldereschi',
                    'fcp_mu': 1e+30,
                    'fixed_magnetization(i), i=1,3': 1e+30,
                    'force_symmorphic': False,
                    'ibrav': 1e+30,
                    'input_dft': 'readfrompseudopotentialfiles',
                    'lambda': 1e+30,
                    'lda_plus_u': True,
                    'lda_plus_u_kind': 0,
                    'lforcet': True,
                    'localization_thr': 0.0,
                    'london': False,
                    'london_c6(i), i=1,ntyp': 1e+30,
                    'london_rcut': 200.0,
                    'london_rvdw(i), i=1,ntyp': 1e+30,
                    'london_s6': 0.75,
                    'lspinorb': True,
                    'nat': 1e+30,
                    'nbnd': 1e+30,
                    'no_t_rev': False,
                    'noinv': False,
                    'noncolin': False,
                    'nosym': False,
                    'nosym_evc': False,
                    'nqx1, nqx2, nqx3': 1e+30,
                    'nr1, nr2, nr3': 1e+30,
                    'nr1s, nr2s, nr3s': 1e+30,
                    'nspin': 1,
                    'ntyp': 1e+30,
                    'occupations': '',
                    'one_atom_occupations': False,
                    'origin_choice': 1,
                    'q2sigma': 0.1,
                    'qcutz': 0.0,
                    'relaxz': False,
                    'report': -1,
                    'rhombohedral': True,
                    'screening_parameter': 0.106,
                    'smearing': 'gaussian',
                    'space_group': 0,
                    'starting_charge(i), i=1,ntyp': 0.0,
                    'starting_magnetization(i), i=1,ntyp': 1e+30,
                    'starting_ns_eigenvalue(m,ispin,ityp), (m,ispin,ityp) = (1,1,1) ... (2*lmax+1,nspin or npol,ntyp)': 1e+30,
                    'starting_spin_angle': False,
                    'tot_charge': 0.0,
                    'tot_magnetization': 1e+30,
                    'ts_vdw_econv_thr': 1e-06,
                    'ts_vdw_isolated': False,
                    'uniqueb': False,
                    'use_all_frac': False,
                    'vdw_corr': '',
                    'x_gamma_extrapolation': True,
                    'xdm': False,
                    'xdm_a1': 0.6836,
                    'xdm_a2': 1.5045,
                    'zgate': 0.5},
        'ATOMIC_FORCES': {'X': '', 'fx, fy, fz': 1e+30},
        'ATOMIC_POSITIONS': {"Card's options:": '(DEPRECATED)alat',
                             'X': '',
                             'if_pos(1), if_pos(2), if_pos(3)': 1,
                             'x, y, z': 1e+30},
        'ATOMIC_SPECIES': {'Mass_X': 1e+30, 'PseudoPot_X': '', 'X': ''},
        'ATOMIC_VELOCITIES': {'V': '', 'vx, vy, vz': 1e+30},
        'CELL_PARAMETERS': {"Card's options:": '', 'v1, v2, v3': 1e+30},
        'CONSTRAINTS': {'constr(1), constr(2), constr(3), constr(4)': '',
                        'constr_target': 1e+30,
                        'constr_tol': 1e+30,
                        'constr_type': '',
                        'nconstr': 1e+30},
        'K_POINTS': {"Card's options:": 'tbipa',
                     'nk1, nk2, nk3': 1e+30,
                     'nks': 1e+30,
                     'sk1, sk2, sk3': 1e+30,
                     'xk_x, xk_y, xk_z, wk': 1e+30},
        'OCCUPATIONS': {'f_inp1': 1e+30, 'f_inp2': 1e+30}}}



import os


class Queue():
    def __init__(self, package = None, parallel = None, queue = None, command = None, type = 'slurm') -> None:
        if command is None:
            command = os.environ.get('ASE_ESPRESSO_COMMAND')
        if 'PACKAGE' in command:
            if 'pw' in package:
                command = command.replace('PACKAGE', package, 1)
                command = command.replace('PACKAGE', 'pw', 2)
            else:
                command = command.replace('PACKAGE', package)
        if 'PREFIX' in command:
            command = command.replace('PREFIX', self.prefix)
        if 'PARALLEL' in command:
            command = command.replace('PARALLEL', parallel)
        pass

    def set_queue(self):
        '''
        If queue, change command to "sbatch .job_file".
        The queue information are written into '.job_file'
        '''
        
        if self.queue:
            script = ''
            if 'config' in self.queue:
                cf = os.path.join(os.environ['HOME'], self.queue['config'])
                file = open(cf, 'r')
                script = file.read()
                # del self.queue['config']
            else:
                for cf in config_files:
                    if os.path.exists(cf):
                        file = open(cf, 'r')
                        script = file.read()
            jobname = self.prefix
            with open('%s/.job_file' % self.directory, 'w') as fh:
                fh.writelines("#!/bin/bash\n")
                fh.writelines("#SBATCH --job-name=%s \n" % jobname)
                fh.writelines("#SBATCH --output=%s.out\n" % self.prefix)
                fh.writelines("#SBATCH --error=%s.err\n" % self.prefix)
                fh.writelines("#SBATCH --wait\n")
                for key, value in self.queue.items():
                    if key == 'config': continue
                    if value:
                        fh.writelines("#SBATCH --%s=%s\n" %(key, value))
                fh.writelines("%s \n"%script)
                fh.writelines("%s \n" % command)
            self.command = "sbatch {0}".format('.job_file')
        else:
            self.command = command
        self.logger.debug('Command: %s'%(self.command))


import xml.etree.ElementTree as ET
from xmlschema import XMLSchema


xmlfile = '/home/xing/ase/xespresso/restart/scf/fe+u/fe+u.save/data-file-schema.xml'

xsd = XMLSchema('/home/xing/ase/xespresso/restart/scf/fe+u/fe+u.save/data-file-schema.xml')
# Validate XML document against the schema
# Returned dictionary has a structure where, if tag ['key'] is "simple", xml_dictionary['key'] returns its content.
# Otherwise, the following keys are available:
#
#  xml_dictionary['key']['$'] returns its content
#  xml_dictionary['key']['@attr'] returns its attribute 'attr'
#  xml_dictionary['key']['nested_key'] goes one level deeper.
xml_dictionary, errors = xsd.to_dict(xmlfile, validation='lax')
xml_version = xml_dictionary['general_info']['xml_format']['@VERSION']
inputs = xml_dictionary.get('input', {})
outputs = xml_dictionary['output']

print(inputs)



from ase import Atoms
from ase import io
from ase.build import bulk, sort
from ase.io import read
from xespresso import Espresso
from itertools import chain
import numpy as np
import copy
import os, shutil
import sys

class HpXEspresso(Espresso):
    def __init__(self, restart=None, ignore_bad_restart_file=False,
                 label='xespresso', atoms=None, package = 'pw', parallel =  '',
                 queue = None,queue_hp = False, parallel_hp = '', iteration=1,
                 sc_type ="relax", relax_type = "vc-relax",
                 insulator = None, magnetic = None, nbnd =  None,
                 parallelization_hp = False,ethr_relax = None, ethr_scf = None, ethr_scf2 = None,
                 **kwargs):
        """
        This class allows to perform the chain of relax-scf-hp calculations necessary to compute U/V values
        via DFPT "Hubbard parameters from density-functional perturbation theory", 
        Phys. Rev. B 98, 085127 (2018); arXiv:1805.01805 using ase and XEspresso, 
        of which XpXEspresso is a subclass.
        
        input_data, pseudopotentials, kspacing, kpts, koffset
            Please have a look at Espresso module in ASE
            
        queue: dict
            A dictionary with parameters for job submission of the pw.x calculation, e.g.
             queue = {'nodes': 4, 'ntasks-per-node': 20, 
                      'account': 'xxx', 'partition': 'normal', 
                      'time': '23:59:00'}
                      
        queue_hp: dict
            A dictionary with parameters for job submission of the hp.x calculation, e.g.
             queue = {'nodes': 4, 'ntasks-per-node': 20, 
                      'account': 'xxx', 'partition': 'normal', 
                      'time': '23:59:00'}
        package: str
            Choose the quantum espresso pacakge: pw, dos, projwfc, band, pp, ph, ..
            For NEB calculation, please use neb.NEBEspresso module.
            Calculaiton use phonon is not implemented yet.
            
        parallel: str
            A str which control the parallelization parameters for the pw.x runs: -nimage, -npools, 
            -nband, -ntg, -ndiag or -northo (shorthands, respectively: 
            -ni, -nk, -nb, -nt, -nd).
        
        parallel_hp: str
            A str which control the parallelization parameters for the hp.x runs: -nimage, -npools, 
            -nband, -ntg, -ndiag or -northo (shorthands, respectively: 
            -ni, -nk, -nb, -nt, -nd).
            
        max_iter: int
            An int indicating the interation step in case a self-consistent (site-dependent) calculation
            of the U/V parameters is performed (Chech the Uscsd module).
        
        sc_type: str 
            A str indicating if the procedure for the self-consistent (site-dependent) calculation
            of the U/V parameters is performed by relaxing or not the structure. Check also the Uscsd class.
            Possible values are "relax" and "scf".
            
        relax_type: str
            Possible values are "vc-relax" and "relax". Indicate thetype of relaxation run incase sc_type = "relax".
            Check also the Uscsd class.
        
        insulator: bool
            A bool indicating if the material is an insulator (True) or a metal (False).
            It has to be specified when sc_type = "scf" and iteration = 1, otherwhise one could not determine 
            the correct procedure to be perfomed.
            
        magnetic: int
            An int indicating if the material is non-magneticc (1) or  magnetic (2)
            It has to be specified when sc_type = "scf" and iteration = 1, otherwhise one could not determine 
            the correct procedure to be perfomed.
            
        nbnd: int
            An int indicating the number of bands to take into account in the calculations.
            It has to be specified when sc_type = "scf" and iteration = 1, otherwhise one could not determine 
            the correct procedure to be perfomed.
            
        parallelization_hp: bool
            A bool indicating if  parallelization of the hp run over atoms or qpoints needs to be performed.
            parallelization_hp = True
            
        ethr_relax: float
            A float corresponfing to the threshold for the scf convergence in the relaxation step.  
            If not specified it is set to 1.0E-8
        
        ethr_scf: float
            A float corresponfing to the threshold for the scf convergence in the (first) scf step.
            If not specified it is set to 1.0E-10
            
        ethr_scf2: float
            A float corresponfing to the threshold for the scf convergence in the second scf step.
            If not specified it is set to 1.0E-12
            
        !!!!!!!!!!!!!!!!! WARNING !!!!!!!!!!!!!!!!!!!
        Do not forget to provide in the input_data dictionry and two elements:
            "lda_plus_U" : True
            "input_ntyp": { "starting_magnetization" : {Hubbard_site: XX},
                            "Hubbard_U" : {Hubbard_site: XX}}
            
        TO DO:
        - check inputs: sc_type, relax_typ, insulator, magnetic, ndnb
        - implement parallelization_hp = True
        - implement possibility of restart of an interrupted relax run or hp run.
        - check copy parameters.in file in the case of DFT+U+V
            
        """
        super().__init__(restart=None, ignore_bad_restart_file=False,
                 label='xespresso', atoms=None, package = 'pw', parallel = '',
                 queue = None, 
                 **kwargs)
        
        self.calc = None
        
        self.relax_directory = None
        self.relax_parameters = None
        self.relax_results = None
        
        self.scf_directory = None
        self.scf_parameters = None
        self.scf_results = None
        
        self.hp_directory = None
        if 'parameters_hp' not in kwargs:
            kwargs['parameters_hp'] = {}
        self.hp_parameters = kwargs['parameters_hp']
        self.hp_results = None
        
        self.directory = os.path.split(label)[0]
        self.prefix = os.path.split(label)[1]
        self.directory_ref = self.directory
        

        self.queue_hp = queue_hp
        self.parallel_hp = parallel_hp
        self.queue = queue
        self.parallel = parallel
            

        self.iteration = iteration
        
        self.sc_type = sc_type
        self.relax_type = relax_type

        if self.sc_type != None and self.sc_type not in ["relax", "scf"]:
            sys.exit("Please provide a str for the sc_type variable: relax or scf")

        if self.relax_type != None and self.relax_type not in ["relax", "vc-relax"]:
            sys.exit("Please provide a str for the relax_type variable: relax or vc-relax")
        if self.sc_type != None and self.sc_type not in ["relax", "scf"]:
            sys.exit("Please provide a bool for the sc_type variable: relax or scf")

        self.insulator = insulator
        self.magnetic = magnetic
        self.nbnd = nbnd

        if self.insulator != None and bool(self.insulator) == False:
            sys.exit("Please provide a bool for the insulator variable")

        if self.magnetic != None and self.magnetic not in [1,2]:
            sys.exit("Please provide an int for the magnetic  variable: 1 for spin unpolirzed and 2 for spin poolariz")

        if self.nbnd != None and type(self.nbnd) != int:
            sys.exit("Please provide an int for the nbnd variable") 
        
        self.parallelization_hp = parallelization_hp
        
        
        self.ethr_relax = ethr_relax
        self.ethr_scf = ethr_scf
        self.ethr_scf2 = ethr_scf2
            
    def read_results_fromdirectory(self,directory):
        """
        Extension of the read_results metho of the XEspresso class, so that results
        can be read from a directory different from self.directory
        
        directory: str
            A str indicating the the path of the directory from which results have to be read.
        """
        from xespresso.xio import get_atomic_species
        pwo = self.prefix + '.pwo'
        
        if directory == None:
            directory = self.directory
        
        pwos = [file for file in os.listdir(directory) if pwo in file]
        output = None
        for pwo in pwos:
            atomic_species = None
            pwo = os.path.join(directory, pwo)
            # print('read results from: %s'%pwo)
            # output = next(read_espresso_out(pwo))
            atomic_species = get_atomic_species(pwo)
            try:
                output = io.read(pwo)
                atomic_species = get_atomic_species(pwo)
                if atomic_species: output.info['species'] = atomic_species
            except Exception as e:
                print(pwo, e)
                # print('\nread %s failed\n'%pwo)
        if output:
            self.calc = output.calc
            self.results = output.calc.results
            self.results['atoms'] = output
            self.efermi = self.get_fermi_level()
            self.nspins = self.get_number_of_spins()
            self.atoms = output
        else:
            print('\nRead result failed\n') 
        
        
    def is_insulator(self,fermi=None):
        """
        Checking if the system is a insulators
        One can provide the fermi energy as input, otherwise it will be extracted from the pw.pwo file.
        WARNING: The Fermi energy cannot beread from a calculation run using "tot_magnetization"
        
        fermi: float
            A float corresponding to the Fermi energy
        
        Results:
        - set with Bool (True=insulator, False=metal), bandgap value (eV), VBM and CBm values (eV)
        """

        if fermi == None:
            fermi = self.calc.get_fermi_level()
            if fermi == None:
                raise Exception("I cannot get the Fermi energy from the pw.pwo file"
                                "Provide a Fermi energy value.")


        # reorganize the bands, rather than per kpoint, per energy level
        bands = []
        nbands = len(self.calc.get_eigenvalues(kpt=0,spin=0))
        for i in range(nbands):
            bands.append([])
            for k in range(len(self.calc.get_k_point_weights())):
                if self.calc.get_number_of_spins() == 1:#calc.nspins == 1:
                    bands[i].append(self.calc.get_eigenvalues(kpt=k)[i])
                else:
                    bands[i].append(self.calc.get_eigenvalues(kpt=k,spin=0)[i])
                    bands[i].append(self.calc.get_eigenvalues(kpt=k,spin=1)[i])
        bands = np.array(bands)
        max_mins = [(max(i), min(i)) for i in bands]

        # one band is crossed by the fermi energy
        if any(i[1] < fermi and fermi < i[0] for i in max_mins):
            return(False, None, None, None)

        # case of semimetals, fermi energy at the crossing of two bands
        # this will only work if the dirac point is computed!
        elif (any(i[0] == fermi for i in max_mins) and
                  any(i[1] == fermi for i in max_mins)):
            return False, 0., fermi, fermi
        # insulating case
        else:
            # take the max of the band maxima below the fermi energy
            homo = max([i[0] for i in max_mins if i[0] < fermi])
            # take the min of the band minima above the fermi energy
            lumo = min([i[1] for i in max_mins if i[1] > fermi])
            gap = lumo - homo
            if gap <= 0.:
                raise Exception("Something wrong has been implemented. "
                                "Revise the code!")
            return True, gap, homo, lumo
        
    def get_magnetizations(self, getall=False):
        """
        Read the total and absolute magnetizations from the pw.x out file
        """

        start = '!    total energy'
        end = '     convergence has'

        mags = dict()
        mags['total magnetization'] = []
        mags['absolute magnetization'] = []
        
        filename = os.path.join(self.directory, self.prefix + ".pwo")
        with open(filename, 'r') as fobj:
            for line in fobj:
                if start in line:
                    while end not in line:
                        line = next(fobj)
                        for mag in mags.keys():
                            if mag in line:
                                mags[mag].append(float(line.split()[3]))
        if getall:
            return mags
        else:
            return [mags[k][-1] if mags[k] else 0.0 for k in
                    sorted(mags.keys())]
        
    def read_Hubbard_parameters(self,iteration):
        """
        Read the U/V values from the Hubbard_parameters.dat file or from the paramets.in file for DFT+U+V
        """
        
        
        if 'lda_plus_u_kind' not in self.parameters or self.parameters['lda_plus_u_kind'] != 1:
            start = 'site n.'
            hp_U = dict()
            directory = os.path.join(self.directory_ref,"hp_"+str(iteration))
            filename = os.path.join(directory, self.prefix + ".Hubbard_parameters.dat")
            
            print(filename)
            try:
                file = open(filename, 'r')
            except IOError:
                print('There was an error opening the Hubbard.dat file!')
                sys.exit('There was an error opening the Hubbard.dat file!')

            
            with open(filename, 'r') as fobj:
                lines = fobj.readlines()
                nlines = len(lines)
                start = [n  for n, line in enumerate(lines) if 'site n.' in line][0]
                end = [n  for n, line in enumerate(lines) if "=--------------" in line][1] -1
                for line in lines[start+1:end]:
                    l = line.split()
                    hp_U[l[0]] = {}
                    hp_U[l[0]]['label'] = l[2]
                    hp_U[l[0]]['spin'] = l[3]
                    hp_U[l[0]]['type'] = l[4]
                    hp_U[l[0]]['new_label'] = l[5]
                    hp_U[l[0]]['U'] = l[6]

        #elif self.scf_parameters['lda_plus_u_kind'] == 2:
        #    hp_U = None
        #    print("Reading the parametes.in file for DFT+U+V is not implemented yet")
        else:
            hp_U = None
            print("lda_plus_u_kind=1 is not implemented")

        return hp_U
        
    def run_relax(self):
        """
        Running the relaxation step
        """
        
        if self.sc_type == "scf":
            sys.exit("You set sc_type to False. I will not relax the structure")
        
        self.directory = os.path.join(self.directory_ref,"relax_"+str(self.iteration))
        self.set_queue(package = "pw", parallel = self.parallel, queue = self.queue)
        
        if self.relax_parameters != None: # This works if the execution of the script is not interrupted
            self.parameters = copy.deepcopy(self.relax_parameters)
        
        self.parameters["calculation"] = self.relax_type
        self.parameters['verbosity'] = 'high'
        if self.ethr_relax == None:
            self.parameters["conv_thr"] = 1.0E-8
        else:
            self.parameters["conv_thr"] = self.ethr_relax
            
        self.parameters['input_data']['restart_mode'] = 'from_scratch'
        
        if self.iteration > 1:
            if 'lda_plus_u_kind' in self.parameters['input_data'] and self.parameters['input_data']['lda_plus_u_kind'] == 2:
                old_parametersin =  os.path.join(self.directory_ref,"hp_"+str(self.iteration -1),'parameters.in')
                shutil.copy(old_parametersin,
                            self.directory ,symlinks=False, ignore=None)
                self.parameters['input_data']['Hubbard_parameters'] = 'file'

    def run_scf(self):
        """
        Running the (first) scf step 
        """
        #self.set_queue(package = "pw", parallel = self.parallel, queue = self.queue)
        if self.sc_type == "relax":
            #self.read_results()
            if self.iteration == 1:
                self.read_results()
            else:
                previous_run = os.path.join(self.directory_ref,"relax_"+str(self.iteration))
                self.read_results_fromdirectory(directory=previous_run)
                self.relax_directory = previous_run
                
            if not self.relax_directory:
                self.relax_directory = self.directory
            if not self.relax_parameters:
                self.relax_parameters = copy.deepcopy(self.parameters)
            if not self.relax_results:
                self.relax_results = copy.deepcopy(self.results)
                self.efermi = self.get_fermi_level()
            # read new atoms from results and set magmoms
            self.atoms = self.results['atoms']
            self.parameters = copy.deepcopy(self.relax_parameters)
            self.parameters['input_data']['restart_mode'] = 'restart'
            
            
            # create working directory
            self.save_directory_old = os.path.join(self.relax_directory, '%s.save'%self.prefix)
            self.directory = os.path.join(self.directory_ref,"scf_"+str(self.iteration))
            self.save_directory = os.path.join(self.directory, '%s.save'%self.prefix)
            self.label = os.path.join(self.directory, self.prefix)
            self.scf_directory = self.directory
            
            if self.iteration > 1:
                if 'lda_plus_u_kind' in self.parameters['input_data'] and self.parameters['input_data']['lda_plus_u_kind'] == 2:
                    old_parametersin =  os.path.join(self.directory_ref,"hp_"+str(self.iteration -1),'parameters.in')
                    shutil.copy(old_parametersin,
                                self.directory ,symlinks=False, ignore=None)
                    self.parameters['input_data']['Hubbard_parameters'] = 'file'
    


            #if path already exists, remove it before copying with copytree()
            if os.path.exists(self.save_directory):
                shutil.rmtree(self.save_directory)
            
            shutil.copytree(self.save_directory_old,
                            self.save_directory,symlinks=False, ignore=None)
            #self.calculate()
            insulator = self.is_insulator()[0]
            magnetic = self.get_number_of_spins()
            nbnd = len(self.calc.get_eigenvalues(kpt=0))
        else:                                  
            self.directory = os.path.join(self.directory_ref,"scf_"+str(self.iteration))
            self.scf_directory = self.directory
                                                 
            if self.iteration == 1:
                if self.insulator == None:
                    sys.exit("You need to provide a value (True/False) for  'insulator'")
                else:
                    insulator = self.insulator
                if self.magnetic == None:
                    sys.exit("You need to provide a value (True/False) for  'magnetic'")
                else:
                    magnetic = self.magnetic
                if self.nbnd == None:
                    sys.exit("You need to provide a value (True/False) for  'nbnd'")
                else:
                    nbnd = self.nbnd
                    
            else:
                previous_run = os.path.join(self.directory_ref,"scf_"+str(self.iteration-1))
                self.read_results_fromdirectory(directory=previous_run)
                insulator = self.is_insulator()[0]
                magnetic = self.get_number_of_spins()
                nbnd = len(self.calc.get_eigenvalues(kpt=0))
                
       
                
        self.parameters['verbosity'] = 'high'
        self.parameters['input_data']['verbosity'] = 'high'
        self.parameters['calculation'] = "scf"
        
        if insulator == True and magnetic==1:
            self.parameters['input_data']["occupations"] = "fixed"
            #nbnd = len(self.calc.get_eigenvalues(kpt=0))
            self.parameters['input_data']["nbnd"] = nbnd
            if  "degauss" in self.parameters['input_data'].keys():
                del self.parameters['input_data']["degauss"]

            if  "starting_magnetization" in self.parameters['input_data']['input_ntyp'].keys():
                del self.parameters['input_data']['input_ntyp']["starting_magnetization"]
            
            if self.ethr_scf == None:
                self.parameters["conv_thr"] = 1.0E-10
            else:
                self.parameters["conv_thr"] = self.ethr_scf


        elif insulator == False or (insulator == True and magnetic==2):
            self.parameters['input_data']["occupations"] = "smearing"
            self.parameters['input_data']["degauss"] = 0.01
            
            if self.ethr_scf == None:
                self.parameters["conv_thr"] = 1.0E-10
            else:
                self.parameters["conv_thr"] = self.ethr_scf
          
        else:
            if self.sc_type =="relax":
                sys.exit("Case not implemented: is an insulator? {} and is spin polarized? {}".format(self.is_insulator()[0],self.get_spin_polarized()) )
            else:
                sys.exit("Case not implemented: is an insulator? {} and is spin polarized? {}".format(self.insulator,self.magnetic))

        self.set_queue(package = "pw", parallel = self.parallel, queue = self.queue)
    
    def run_scf2(self):
        """
        Running the second scf step (if necessary)
        """
        self.read_results()       
        if self.get_number_of_spins() == 2 and self.is_insulator()[0] == True:

                # save relax parameters
                if not self.scf_directory:
                    self.scf_directory = self.directory
                if not self.scf_parameters:
                    self.scf_parameters = copy.deepcopy(self.parameters)
                if not self.scf_results:
                    self.scf_results = copy.deepcopy(self.results)
                    self.efermi = self.get_fermi_level()

                # read new atoms from results and set scf specific parameters 
                self.atoms = self.results['atoms']
                self.parameters = copy.deepcopy(self.scf_parameters)
                self.parameters['calculation'] = "scf"
                self.parameters['input_data']['verbosity'] = 'high'
                self.parameters['input_data']['restart_mode'] = 'restart'
                self.parameters['input_data']["occupations"] = "fixed"
                nbnd = len(self.calc.get_eigenvalues(kpt=0))
                self.parameters['input_data']["nbnd"] = nbnd
                self.parameters['input_data']["tot_magnetization"] = abs(self.get_magnetizations(True)['total magnetization'][-1])

                if  "degauss" in self.parameters['input_data'].keys():
                    del self.parameters['input_data']["degauss"]
                    
                if  "starting_magnetization" in self.parameters['input_data']['input_ntyp'].keys():
                    del self.parameters['input_data']['input_ntyp']["starting_magnetization"]
                    
                self.parameters['input_data']["startingpot"] = "file"
                self.parameters['input_data']["startingwfc"] = "file"

                if self.ethr_scf2 == None:
                    self.parameters["conv_thr"] = 1.0E-12
                else:
                    self.parameters["conv_thr"] = self.ethr_scf2
                


                # create working directory
                self.save_directory_old = os.path.join(self.scf_directory, '%s.save'%self.prefix)
                self.directory = os.path.join(self.directory_ref,"scf2_"+str(self.iteration))
                self.save_directory = os.path.join(self.directory, '%s.save'%self.prefix)
                self.label = os.path.join(self.directory, self.prefix)
                self.scf_directory = self.directory
                
                #if path already exists, remove it before copying with copytree()
                if os.path.exists(self.save_directory):
                    shutil.rmtree(self.save_directory)

                shutil.copytree(self.save_directory_old,
                                self.save_directory ,symlinks=False, ignore=None)
                
                self.scf2_necessary = True
        else:
            self.scf2_necessary = False
            print("No need to perform a second scf")
            pass
   
        self.set_queue(package = "pw", parallel = self.parallel, queue = self.queue)
            
    def run_hp(self): 
        """
        Running the hp step
        """
        # create working directory
        if self.scf_directory == None:
            if os.path.exists(os.path.join(self.directory_ref,"scf2_"+str(self.iteration))):
                self.scf_directory = os.path.join(self.directory_ref,"scf2_"+str(self.iteration))
            else:
                self.scf_directory = os.path.join(self.directory_ref,"scf_"+str(self.iteration))
          
        if self.iteration == 1:
            self.read_results()  
        else:
            self.read_results_fromdirectory(directory=self.scf_directory)


            
        if self.parallelization_hp == False:
            if not self.hp_parameters:
                self.hp_parameters = copy.deepcopy(self.parameters_hp)
            if not self.hp_results:
                self.hp_results = copy.deepcopy(self.results)

            
            self.save_directory_old = os.path.join(self.scf_directory, '%s.save'%self.prefix)
            self.directory = os.path.join(self.directory_ref,"hp_"+str(self.iteration))
            #self.directory = os.path.join(self.directory_ref,"hp_",str(iteration))
            self.save_directory = os.path.join(self.directory, '%s.save'%self.prefix)
            self.label = os.path.join(self.directory, self.prefix)
            self.hp_directory = self.directory

            #if path already exists, remove it before copying with copytree()
            if os.path.exists(self.save_directory):
                shutil.rmtree(self.save_directory)

            shutil.copytree(self.save_directory_old,
                            self.save_directory ,symlinks=False, ignore=None)

            self.post(package = 'hp',queue = self.queue_hp, parallel = self.parallel_hp, **self.hp_parameters)
        else:
            sys.exit("Parallelization of the hp.x calculation over the atoms or q-points not implemented.")
            
            

